from __future__ import unicode_literals

from itertools import product

from mock import Mock, sentinel, patch, call
from pytest import raises, mark

from concrete.util.redis_io import (
    read_communication_from_redis_key,
    write_communication_to_redis_key,
    RedisCommunicationReader,
    RedisCommunicationWriter,
)


@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_read_communication_from_redis_key(mock_read_communication_from_buffer):
    redis = Mock(get=Mock(side_effect=[sentinel.buf]))
    mock_read_communication_from_buffer.side_effect = [sentinel.comm]
    assert read_communication_from_redis_key(
        redis, sentinel.key, sentinel.add_references) == sentinel.comm
    redis.get.assert_called_once_with(sentinel.key)
    mock_read_communication_from_buffer.assert_called_once_with(
        sentinel.buf, add_references=sentinel.add_references)


@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_read_communication_from_redis_key_empty(mock_read_communication_from_buffer):
    redis = Mock(get=Mock(side_effect=[None]))
    assert read_communication_from_redis_key(
        redis, sentinel.key, sentinel.add_references) is None
    redis.get.assert_called_once_with(sentinel.key)
    assert not mock_read_communication_from_buffer.called


@patch('concrete.util.redis_io.write_communication_to_buffer')
def test_write_communication_to_redis_key(mock_write_communication_to_buffer):
    redis = Mock(set=Mock())
    mock_write_communication_to_buffer.side_effect = [sentinel.buf]
    write_communication_to_redis_key(redis, sentinel.key, sentinel.comm)
    redis.set.assert_called_once_with(sentinel.key, sentinel.buf)
    mock_write_communication_to_buffer.assert_called_once_with(sentinel.comm)


@mark.parametrize(
    'right_to_left,pop,inferred,block',
    list(product((False, True), (False, True), (False, True), (False,))) +
    list(product((False, True), (True,), (False, True), (True,))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_list(
        mock_read_communication_from_buffer, right_to_left, pop, inferred, block):
    redis = Mock(
        llen=Mock(return_value=2),
        lindex=Mock(side_effect=[sentinel.buf0, sentinel.buf1]),
        lpop=Mock(side_effect=[sentinel.buf0, sentinel.buf1, None]),
        rpop=Mock(side_effect=[sentinel.buf0, sentinel.buf1, None]),
        blpop=Mock(side_effect=[
            (sentinel.key, sentinel.buf0), (sentinel.key, sentinel.buf1), None]),
        brpop=Mock(side_effect=[
            (sentinel.key, sentinel.buf0), (sentinel.key, sentinel.buf1), None]),
        type=Mock(return_value=b'list'),
    )
    mock_read_communication_from_buffer.side_effect = [
        sentinel.comm0, sentinel.comm1
    ]

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type=None if inferred else 'list',
        add_references=sentinel.add_references,
        right_to_left=right_to_left, pop=pop, block=block,
        block_timeout=sentinel.block_timeout)

    it = iter(reader)
    assert next(it) == sentinel.comm0
    assert next(it) == sentinel.comm1
    with raises(StopIteration):
        next(it)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    if pop:
        assert not redis.llen.called
        assert not redis.lindex.called
        if block:
            assert not redis.lpop.called
            assert not redis.rpop.called
            getattr(redis, 'brpop' if right_to_left else 'blpop').assert_has_calls([
                call(sentinel.key, timeout=sentinel.block_timeout),
                call(sentinel.key, timeout=sentinel.block_timeout),
                call(sentinel.key, timeout=sentinel.block_timeout),
            ])
            assert not getattr(redis, 'blpop' if right_to_left else 'brpop').called
        else:
            getattr(redis, 'rpop' if right_to_left else 'lpop').assert_has_calls([
                call(sentinel.key),
                call(sentinel.key),
                call(sentinel.key),
            ])
            assert not getattr(redis, 'lpop' if right_to_left else 'rpop').called
            assert not redis.blpop.called
            assert not redis.brpop.called
    else:
        redis.llen.assert_called_with(sentinel.key)
        redis.lindex.assert_has_calls([
            call(sentinel.key, -1 if right_to_left else 0),
            call(sentinel.key, -2 if right_to_left else 1),
        ])
        assert not redis.lpop.called
        assert not redis.rpop.called
        assert not redis.blpop.called
        assert not redis.brpop.called

    mock_read_communication_from_buffer.assert_has_calls([
        call(sentinel.buf0, add_references=sentinel.add_references),
        call(sentinel.buf1, add_references=sentinel.add_references),
    ])


@mark.parametrize(
    'right_to_left,pop,block',
    list(product((False, True), (False, True), (False,))) +
    list(product((False, True), (True,), (True,))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_list_empty(
        mock_read_communication_from_buffer, right_to_left, pop, block):
    redis = Mock(
        llen=Mock(return_value=0),
        lindex=Mock(),
        lpop=Mock(side_effect=[None]),
        rpop=Mock(side_effect=[None]),
        blpop=Mock(side_effect=[None]),
        brpop=Mock(side_effect=[None]),
    )

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type='list',
        add_references=sentinel.add_references,
        right_to_left=right_to_left, pop=pop, block=block,
        block_timeout=sentinel.block_timeout)

    it = iter(reader)
    with raises(StopIteration):
        next(it)

    if pop:
        if block:
            assert not redis.llen.called
            assert not redis.lindex.called
            assert not redis.lpop.called
            assert not redis.rpop.called
            getattr(redis, 'brpop' if right_to_left else 'blpop').assert_called_once_with(
                sentinel.key, timeout=sentinel.block_timeout)
            assert not getattr(redis, 'blpop' if right_to_left else 'brpop').called
        else:
            assert not redis.llen.called
            assert not redis.lindex.called
            getattr(redis, 'rpop' if right_to_left else 'lpop').assert_called_once_with(
                sentinel.key)
            assert not getattr(redis, 'lpop' if right_to_left else 'rpop').called
            assert not redis.blpop.called
            assert not redis.brpop.called
    else:
        redis.llen.assert_called_with(sentinel.key)
        assert not redis.lindex.called
        assert not redis.lpop.called
        assert not redis.rpop.called
        assert not redis.blpop.called
        assert not redis.brpop.called

    assert not mock_read_communication_from_buffer.called


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_set(
        mock_read_communication_from_buffer, right_to_left, inferred):

    def _sadd(_, buf):
        return {
            sentinel.buf0: 1,
            sentinel.buf1: 0,
            sentinel.buf2: 0,
            sentinel.buf3: 1,
            sentinel.buf4: 0,
        }[buf]

    redis = Mock(
        sscan=Mock(side_effect=[
            (7, [sentinel.buf0, sentinel.buf1]),
            (3, [sentinel.buf2]),
            (0, [sentinel.buf3, sentinel.buf4]),
        ]),
        sadd=Mock(side_effect=_sadd),
        scard=Mock(return_value=2),
        expire=Mock(),
        exists=Mock(side_effect=[True, False]),
        type=Mock(return_value=b'set'),
    )
    mock_read_communication_from_buffer.side_effect = [
        sentinel.comm0, sentinel.comm3
    ]

    reader = RedisCommunicationReader(
        redis, 'my-key', key_type=None if inferred else 'set',
        add_references=sentinel.add_references,
        right_to_left=right_to_left)

    it = iter(reader)
    assert next(it) == sentinel.comm0
    assert next(it) == sentinel.comm3
    with raises(StopIteration):
        next(it)

    if inferred:
        redis.type.assert_called_once_with('my-key')
    else:
        assert not redis.type.called

    redis.sscan.assert_has_calls([
        call('my-key', 0),
        call('my-key', 7),
        call('my-key', 3),
    ])

    mock_read_communication_from_buffer.assert_has_calls([
        call(sentinel.buf0, add_references=sentinel.add_references),
        call(sentinel.buf3, add_references=sentinel.add_references),
    ])


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_set_batch(
        mock_read_communication_from_buffer, right_to_left, inferred):
    redis = Mock(
        srandmember=Mock(side_effect=[
            [sentinel.buf0, sentinel.buf1],
        ]),
        type=Mock(return_value=b'set'),
    )
    mock_read_communication_from_buffer.side_effect = [
        sentinel.comm0, sentinel.comm1
    ]

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type=None if inferred else 'set',
        add_references=sentinel.add_references,
        right_to_left=right_to_left)

    assert reader.batch(sentinel.batch_size) == [sentinel.comm0, sentinel.comm1]

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    redis.srandmember.assert_called_once_with(sentinel.key, sentinel.batch_size)

    mock_read_communication_from_buffer.assert_has_calls([
        call(sentinel.buf0, add_references=sentinel.add_references),
        call(sentinel.buf1, add_references=sentinel.add_references),
    ])


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_hash(
        mock_read_communication_from_buffer, right_to_left, inferred):

    def _sadd(_, comm_id):
        return {
            sentinel.comm_id0: 1,
            sentinel.comm_id1: 0,
            sentinel.comm_id2: 0,
            sentinel.comm_id3: 1,
            sentinel.comm_id4: 0,
        }[comm_id]

    redis = Mock(
        hscan=Mock(side_effect=[
            (7, {sentinel.comm_id0: sentinel.buf0, sentinel.comm_id1: sentinel.buf1}),
            (3, {sentinel.comm_id2: sentinel.buf2}),
            (0, {sentinel.comm_id3: sentinel.buf3, sentinel.comm_id4: sentinel.buf4}),
        ]),
        sadd=Mock(side_effect=_sadd),
        hlen=Mock(return_value=2),
        expire=Mock(),
        exists=Mock(side_effect=[True, False]),
        type=Mock(return_value=b'hash'),
    )
    mock_read_communication_from_buffer.side_effect = [
        sentinel.comm0, sentinel.comm3
    ]

    reader = RedisCommunicationReader(
        redis, 'my-key', key_type=None if inferred else 'hash',
        add_references=sentinel.add_references,
        right_to_left=right_to_left)

    it = iter(reader)
    assert next(it) == sentinel.comm0
    assert next(it) == sentinel.comm3
    with raises(StopIteration):
        next(it)

    if inferred:
        redis.type.assert_called_once_with('my-key')
    else:
        assert not redis.type.called

    redis.hscan.assert_has_calls([
        call('my-key', 0),
        call('my-key', 7),
        call('my-key', 3),
    ])

    mock_read_communication_from_buffer.assert_has_calls([
        call(sentinel.buf0, add_references=sentinel.add_references),
        call(sentinel.buf3, add_references=sentinel.add_references),
    ])


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_set_pop(
        mock_read_communication_from_buffer, right_to_left, inferred):
    redis = Mock(
        spop=Mock(side_effect=[sentinel.buf0, sentinel.buf1, None]),
        type=Mock(return_value=b'set'),
    )
    mock_read_communication_from_buffer.side_effect = [
        sentinel.comm0, sentinel.comm1
    ]

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type=None if inferred else 'set',
        add_references=sentinel.add_references,
        right_to_left=right_to_left, pop=True)

    it = iter(reader)
    assert next(it) == sentinel.comm0
    assert next(it) == sentinel.comm1
    with raises(StopIteration):
        next(it)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    redis.spop.assert_has_calls([
        call(sentinel.key),
        call(sentinel.key),
        call(sentinel.key),
    ])

    mock_read_communication_from_buffer.assert_has_calls([
        call(sentinel.buf0, add_references=sentinel.add_references),
        call(sentinel.buf1, add_references=sentinel.add_references),
    ])


@mark.parametrize(
    'right_to_left',
    [(False,), (True,)]
)
@patch('concrete.util.redis_io.read_communication_from_buffer')
def test_redis_communication_reader_set_pop_empty(
        mock_read_communication_from_buffer, right_to_left):
    redis = Mock(
        spop=Mock(side_effect=[None]),
        type=Mock(return_value=b'set'),
    )

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type='set',
        add_references=sentinel.add_references,
        right_to_left=right_to_left, pop=True)

    it = iter(reader)
    with raises(StopIteration):
        next(it)

    assert not redis.type.called

    redis.spop.assert_called_once_with(sentinel.key)

    assert not mock_read_communication_from_buffer.called


@mark.parametrize(
    'right_to_left,inferred,key_type,pop,block',
    list(product((False, True), (False, True), ('set', 'list', 'hash'), (False,), (False,))) +
    list(product((False, True), (False, True), ('set', 'list'), (True,), (False,))) +
    list(product((False, True), (False, True), ('list',), (True,), (True,))),
)
def test_redis_communication_reader_len(right_to_left, inferred, key_type, pop, block):
    redis = Mock(
        llen=Mock(side_effect=[3]),
        hlen=Mock(side_effect=[3]),
        scard=Mock(side_effect=[3]),
        type=Mock(return_value=key_type.encode('utf-8')),
    )

    reader = RedisCommunicationReader(
        redis, sentinel.key, key_type=None if inferred else key_type,
        right_to_left=right_to_left, pop=pop, block=block)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    len(reader)

    if key_type == 'set':
        assert redis.scard.called_once_with(sentinel.key)
        assert not redis.llen.called
        assert not redis.hlen.called
    elif key_type == 'list':
        assert not redis.scard.called
        assert redis.llen.called_once_with(sentinel.key)
        assert not redis.hlen.called
    elif key_type == 'hash':
        assert not redis.scard.called
        assert not redis.llen.called
        assert redis.hlen.called_once_with(sentinel.key)
    else:
        raise ValueError('unexpected key type {}'.format(key_type))


def test_redis_communication_reader_failed_infer():
    redis = Mock(
        type=Mock(return_value=b'none'),
    )
    with raises(Exception):
        RedisCommunicationReader(
            redis, sentinel.key,
            add_references=sentinel.add_references)
    redis.type.assert_called_with(sentinel.key)


@mark.parametrize(
    'key_type',
    [('string',), ('zset',)]
)
def test_redis_communication_reader_failed_key_type(key_type):
    with raises(ValueError):
        RedisCommunicationReader(
            sentinel.redis, sentinel.key, key_type=key_type)


@mark.parametrize(
    'right_to_left,inferred,key_type,pop',
    list(product((False, True), (False, True), ('set', 'hash'), (True,))) +
    list(product((False, True), (False, True), ('set', 'hash', 'list'), (False,))),
)
def test_redis_communication_reader_failed_block(right_to_left, inferred, key_type, pop):
    redis = Mock(
        type=Mock(return_value=key_type.encode('utf-8')),
    )

    with raises(ValueError):
        RedisCommunicationReader(
            redis, sentinel.key, key_type=None if inferred else key_type,
            add_references=sentinel.add_references,
            pop=pop, block=True)

    if not inferred:
        assert not redis.type.called


@mark.parametrize(
    'right_to_left,inferred,key_type,pop,block',
    list(product((False, True), (False, True), ('set',), (False, True), (False,))) +
    list(product((False, True), (False, True), ('hash',), (False,), (False,))) +
    list(product((False, True), (False, True), ('list',), (False, True), (False,))) +
    list(product((False, True), (False, True), ('list',), (True,), (True,))),
)
def test_redis_communication_reader_failed_batch(right_to_left, inferred,
                                                 key_type, pop, block):
    redis = Mock(
        type=Mock(return_value=key_type.encode('utf-8')),
    )

    reader = RedisCommunicationReader(
            redis, sentinel.key, key_type=None if inferred else key_type,
            add_references=sentinel.add_references,
            pop=pop, block=block)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    with raises(Exception):
        reader.batch(sentinel.batch_size)


@mark.parametrize(
    'right_to_left,inferred,block',
    list(product((False, True), (False, True), (False, True))),
)
def test_redis_communication_reader_failed_pop(right_to_left, inferred, block):
    redis = Mock(
        type=Mock(return_value=b'hash'),
    )

    with raises(ValueError):
        RedisCommunicationReader(
            redis, sentinel.key, key_type=None if inferred else 'hash',
            add_references=sentinel.add_references,
            block=block, pop=True)

    if not inferred:
        assert not redis.type.called


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.write_communication_to_buffer')
def test_redis_communication_writer_list(
        mock_write_communication_to_buffer, right_to_left, inferred):
    redis = Mock(
        lpush=Mock(side_effect=[3]),
        rpush=Mock(side_effect=[3]),
        type=Mock(return_value=b'list'),
    )
    mock_write_communication_to_buffer.side_effect = [sentinel.buf]

    writer = RedisCommunicationWriter(
        redis, sentinel.key, key_type=None if inferred else 'list',
        right_to_left=right_to_left)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    writer.write(sentinel.comm)

    getattr(redis, 'lpush' if right_to_left else 'rpush').assert_called_once_with(
        sentinel.key, sentinel.buf)
    assert not getattr(redis, 'rpush' if right_to_left else 'lpush').called

    mock_write_communication_to_buffer.assert_called_once_with(sentinel.comm)


@mark.parametrize(
    'right_to_left,inferred',
    list(product((False, True), (False, True))),
)
@patch('concrete.util.redis_io.write_communication_to_buffer')
def test_redis_communication_writer_set(
        mock_write_communication_to_buffer, right_to_left, inferred):
    redis = Mock(
        sadd=Mock(side_effect=[0]),
        type=Mock(return_value=b'set'),
    )
    mock_write_communication_to_buffer.side_effect = [sentinel.buf]

    writer = RedisCommunicationWriter(
        redis, sentinel.key, key_type=None if inferred else 'set',
        right_to_left=right_to_left)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    writer.write(sentinel.comm)

    redis.sadd.assert_called_once_with(sentinel.key, sentinel.buf)
    assert not getattr(redis, 'rpush' if right_to_left else 'lpush').called

    mock_write_communication_to_buffer.assert_called_once_with(sentinel.comm)


@mark.parametrize(
    'right_to_left,inferred,uuid_hash_key',
    list(product((False, True), (False, True), (False, True))),
)
@patch('concrete.util.redis_io.write_communication_to_buffer')
def test_redis_communication_writer_hash(
        mock_write_communication_to_buffer, right_to_left, inferred,
        uuid_hash_key):
    redis = Mock(
        hset=Mock(side_effect=[0]),
        type=Mock(return_value=b'hash'),
    )
    comm = Mock(id=sentinel.comm_id, uuid=Mock(uuidString=sentinel.comm_uuid))
    mock_write_communication_to_buffer.side_effect = [sentinel.buf]

    writer = RedisCommunicationWriter(
        redis, sentinel.key, key_type=None if inferred else 'hash',
        right_to_left=right_to_left, uuid_hash_key=uuid_hash_key)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    writer.write(comm)

    redis.hset.assert_called_once_with(
        sentinel.key,
        sentinel.comm_uuid if uuid_hash_key else sentinel.comm_id,
        sentinel.buf)
    assert not getattr(redis, 'rpush' if right_to_left else 'lpush').called

    mock_write_communication_to_buffer.assert_called_once_with(comm)


@mark.parametrize(
    'right_to_left,inferred,key_type',
    list(product((False, True), (False, True), ('set', 'list', 'hash'))),
)
def test_redis_communication_writer_clear(right_to_left, inferred, key_type):
    redis = Mock(
        delete=Mock(side_effect=[0]),
        type=Mock(return_value=key_type.encode('utf-8')),
    )

    writer = RedisCommunicationWriter(
        redis, sentinel.key, key_type=None if inferred else 'hash',
        right_to_left=right_to_left)

    if inferred:
        redis.type.assert_called_once_with(sentinel.key)
    else:
        assert not redis.type.called

    writer.clear()

    redis.delete.assert_called_once_with(sentinel.key)


def test_redis_communication_writer_failed_infer():
    redis = Mock(
        type=Mock(return_value=b'none'),
    )
    with raises(Exception):
        RedisCommunicationWriter(
            redis, sentinel.key)
    redis.type.assert_called_with(sentinel.key)


@mark.parametrize(
    'key_type',
    [('string',), ('zset',)]
)
def test_redis_communication_writer_failed_key_type(key_type):
    with raises(ValueError):
        RedisCommunicationWriter(
            sentinel.redis, sentinel.key, key_type=key_type)



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.services.ttypes
import concrete.uuid.ttypes
import concrete.communication.ttypes

from thrift.transport import TTransport
all_structs = []


class AnnotationTask(object):
    """
    Annotation task including information for pulling data.

    Attributes:
     - type: Type of annotation task
     - language: Language of the data for the task
     - unitType: Entire communication or individual sentences
     - units: Identifiers for each annotation unit
    """


    def __init__(self, type=None, language=None, unitType=None, units=None,):
        self.type = type
        self.language = language
        self.unitType = unitType
        self.units = units

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.language = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.unitType = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.units = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = concrete.services.ttypes.AnnotationUnitIdentifier()
                        _elem5.read(iprot)
                        self.units.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('AnnotationTask')
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 1)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        if self.language is not None:
            oprot.writeFieldBegin('language', TType.STRING, 2)
            oprot.writeString(self.language.encode('utf-8') if sys.version_info[0] == 2 else self.language)
            oprot.writeFieldEnd()
        if self.unitType is not None:
            oprot.writeFieldBegin('unitType', TType.I32, 3)
            oprot.writeI32(self.unitType)
            oprot.writeFieldEnd()
        if self.units is not None:
            oprot.writeFieldBegin('units', TType.LIST, 4)
            oprot.writeListBegin(TType.STRUCT, len(self.units))
            for iter6 in self.units:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.type is None:
            raise TProtocolException(message='Required field type is unset!')
        if self.unitType is None:
            raise TProtocolException(message='Required field unitType is unset!')
        if self.units is None:
            raise TProtocolException(message='Required field units is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Annotation(object):
    """
    Annotation on a communication.

    Attributes:
     - id: Identifier of the part of the communication being annotated.
     - communication: Communication with the annotation stored in it.
    The location of the annotation depends on the annotation unit identifier
    """


    def __init__(self, id=None, communication=None,):
        self.id = id
        self.communication = communication

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.id = concrete.services.ttypes.AnnotationUnitIdentifier()
                    self.id.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.communication = concrete.communication.ttypes.Communication()
                    self.communication.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Annotation')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRUCT, 1)
            self.id.write(oprot)
            oprot.writeFieldEnd()
        if self.communication is not None:
            oprot.writeFieldBegin('communication', TType.STRUCT, 2)
            self.communication.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.communication is None:
            raise TProtocolException(message='Required field communication is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(AnnotationTask)
AnnotationTask.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'type', None, None, ),  # 1
    (2, TType.STRING, 'language', 'UTF8', None, ),  # 2
    (3, TType.I32, 'unitType', None, None, ),  # 3
    (4, TType.LIST, 'units', (TType.STRUCT, [concrete.services.ttypes.AnnotationUnitIdentifier, None], False), None, ),  # 4
)
all_structs.append(Annotation)
Annotation.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'id', [concrete.services.ttypes.AnnotationUnitIdentifier, None], None, ),  # 1
    (2, TType.STRUCT, 'communication', [concrete.communication.ttypes.Communication, None], None, ),  # 2
)
fix_spec(all_structs)
del all_structs



from __future__ import unicode_literals
from concrete.util import (
    DirectoryBackedCommunicationContainer,
    MemoryBackedCommunicationContainer,
    ZipFileBackedCommunicationContainer,
    RedisHashBackedCommunicationContainer,
    S3BackedCommunicationContainer,
)
from concrete.util import create_comm
from concrete.util import write_communication_to_buffer

from concrete.validate import validate_communication

from pytest import raises, fixture
from mock import Mock, sentinel, patch, call


@fixture
def comm_id_and_buf(request):
    comm_id = 'temp comm'
    return (comm_id, write_communication_to_buffer(create_comm(comm_id)))


def test_directory_backed_comm_container_find_files_recursively():
    directory_path = u'tests/testdata/a'
    cc = DirectoryBackedCommunicationContainer(directory_path)
    assert 3 == len(cc)


def test_directory_backed_comm_container_retrieve():
    directory_path = u'tests/testdata/a'
    cc = DirectoryBackedCommunicationContainer(directory_path)
    assert 3 == len(cc)
    assert u'simple_1' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_memory_backed_comm_container_file_too_large():
    comm_path = u'tests/testdata/simple.tar.gz'
    with raises(Exception):
        MemoryBackedCommunicationContainer(comm_path, max_file_size=500)


def test_memory_backed_comm_container_retrieve():
    comm_path = u'tests/testdata/simple.tar.gz'
    cc = MemoryBackedCommunicationContainer(comm_path)
    assert 3 == len(cc)
    assert u'one' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_zip_file_backed_comm_container_retrieve():
    zipfile_path = u'tests/testdata/simple.zip'
    cc = ZipFileBackedCommunicationContainer(zipfile_path)
    assert 3 == len(cc)
    assert u'simple_1' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_redis_hash_backed_comm_container_iter():
    redis_db = Mock(hkeys=Mock(side_effect=[[
        sentinel.name0, sentinel.name1, sentinel.name2
    ]]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    it = iter(cc)
    assert next(it) == sentinel.name0
    assert next(it) == sentinel.name1
    assert next(it) == sentinel.name2
    with raises(StopIteration):
        next(it)
    redis_db.hkeys.assert_called_once_with(key)


def test_redis_hash_backed_comm_container_len():
    redis_db = Mock(hlen=Mock(side_effect=[3]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    assert 3 == len(cc)
    redis_db.hlen.assert_called_once_with(key)


def test_redis_hash_backed_comm_container_not_contains():
    redis_db = Mock(hexists=Mock(side_effect=[False]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    assert sentinel.comm_id not in cc
    redis_db.hexists.assert_called_once_with(key, sentinel.comm_id)


def test_redis_hash_backed_comm_container_contains():
    redis_db = Mock(hexists=Mock(side_effect=[True]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    assert sentinel.comm_id in cc
    redis_db.hexists.assert_called_once_with(key, sentinel.comm_id)


@patch('concrete.util.comm_container.read_communication_from_buffer')
def test_redis_hash_backed_comm_container_not_getitem(
        mock_read_communication_from_buffer):
    redis_db = Mock(hget=Mock(side_effect=[None]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    with raises(KeyError):
        cc[sentinel.comm_id]
    redis_db.hget.assert_called_once_with(key, sentinel.comm_id)
    assert not mock_read_communication_from_buffer.called


@patch('concrete.util.comm_container.read_communication_from_buffer')
def test_redis_hash_backed_comm_container_getitem(
        mock_read_communication_from_buffer):
    redis_db = Mock(hget=Mock(side_effect=[sentinel.comm_buf]))
    key = sentinel.key
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    mock_read_communication_from_buffer.side_effect = [sentinel.comm]
    assert cc[sentinel.comm_id] == sentinel.comm
    redis_db.hget.assert_called_once_with(key, sentinel.comm_id)
    mock_read_communication_from_buffer.assert_called_once_with(sentinel.comm_buf)


@patch('concrete.util.comm_container.unprefix_s3_key')
def test_s3_backed_comm_container_iter(mock_unprefix_s3_key):
    keys = [Mock(), Mock(), Mock()]
    keys[0].name = sentinel.prefixed_name0
    keys[1].name = sentinel.prefixed_name1
    keys[2].name = sentinel.prefixed_name2
    bucket = Mock(list=Mock(side_effect=[keys]))
    mock_unprefix_s3_key.side_effect = [
        sentinel.name0,
        sentinel.name1,
        sentinel.name2,
    ]
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    it = iter(cc)
    assert next(it) == sentinel.name0
    assert next(it) == sentinel.name1
    assert next(it) == sentinel.name2
    with raises(StopIteration):
        next(it)
    bucket.list.assert_called_once_with()
    mock_unprefix_s3_key.assert_has_calls([
        call(sentinel.prefixed_name0, sentinel.prefix_len),
        call(sentinel.prefixed_name1, sentinel.prefix_len),
        call(sentinel.prefixed_name2, sentinel.prefix_len),
    ])


@patch('concrete.util.comm_container.unprefix_s3_key')
def test_s3_backed_comm_container_len(mock_unprefix_s3_key):
    keys = [Mock(), Mock(), Mock()]
    keys[0].name = sentinel.prefixed_name0
    keys[1].name = sentinel.prefixed_name1
    keys[2].name = sentinel.prefixed_name2
    bucket = Mock(list=Mock(side_effect=[keys]))
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    assert 3 == len(cc)
    bucket.list.assert_called_once_with()


@patch('concrete.util.comm_container.prefix_s3_key')
def test_s3_backed_comm_container_not_contains(mock_prefix_s3_key):
    bucket = Mock(get_key=Mock(side_effect=[None]))
    mock_prefix_s3_key.side_effect = [sentinel.prefixed_name]
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    assert sentinel.comm_id not in cc
    bucket.get_key.assert_called_once_with(sentinel.prefixed_name)
    mock_prefix_s3_key.assert_called_once_with(sentinel.comm_id, sentinel.prefix_len)


@patch('concrete.util.comm_container.prefix_s3_key')
def test_s3_backed_comm_container_contains(mock_prefix_s3_key):
    bucket = Mock(get_key=Mock(side_effect=[sentinel.key]))
    mock_prefix_s3_key.side_effect = [sentinel.prefixed_name]
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    assert sentinel.comm_id in cc
    bucket.get_key.assert_called_once_with(sentinel.prefixed_name)
    mock_prefix_s3_key.assert_called_once_with(sentinel.comm_id, sentinel.prefix_len)


@patch('concrete.util.comm_container.prefix_s3_key')
@patch('concrete.util.comm_container.read_communication_from_buffer')
def test_s3_backed_comm_container_not_getitem(
        mock_read_communication_from_buffer, mock_prefix_s3_key):
    bucket = Mock(get_key=Mock(side_effect=[None]))
    mock_prefix_s3_key.side_effect = [sentinel.prefixed_name]
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    with raises(KeyError):
        cc[sentinel.comm_id]
    bucket.get_key.assert_called_once_with(sentinel.prefixed_name)
    mock_prefix_s3_key.assert_called_once_with(sentinel.comm_id, sentinel.prefix_len)
    assert not mock_read_communication_from_buffer.called


@patch('concrete.util.comm_container.prefix_s3_key')
@patch('concrete.util.comm_container.read_communication_from_buffer')
def test_s3_backed_comm_container_getitem(
        mock_read_communication_from_buffer, mock_prefix_s3_key):
    bucket = Mock(get_key=Mock(side_effect=[
        Mock(get_contents_as_string=Mock(side_effect=[sentinel.comm_buf])),
    ]))
    mock_prefix_s3_key.side_effect = [sentinel.prefixed_name]
    mock_read_communication_from_buffer.side_effect = [sentinel.comm]
    cc = S3BackedCommunicationContainer(bucket, sentinel.prefix_len)

    assert cc[sentinel.comm_id] == sentinel.comm
    bucket.get_key.assert_called_once_with(sentinel.prefixed_name)
    mock_prefix_s3_key.assert_called_once_with(sentinel.comm_id, sentinel.prefix_len)
    mock_read_communication_from_buffer.assert_called_once_with(sentinel.comm_buf)



from __future__ import unicode_literals
from pytest import fixture
from concrete.util import (
    CommunicationReader, FileType
)
import os
import tempfile

from multiprocessing import Process

from pytest import mark


@fixture
def fifo(request):
    (fd, path) = tempfile.mkstemp()
    os.close(fd)
    os.remove(path)
    os.mkfifo(path)  # not quite secure...

    def _remove():
        if os.path.exists(path):
            os.remove(path)
    request.addfinalizer(_remove)

    return path


def write_fifo(input_path, fifo_path):
    with open(input_path, 'rb') as in_f:
        with open(fifo_path, 'wb') as out_f:
            out_f.write(in_f.read())


@mark.posix
def test_concatenated(fifo):
    input_path = 'tests/testdata/simple_concatenated'
    p = Process(target=write_fifo, args=(input_path, fifo))
    p.start()

    reader = CommunicationReader(fifo, filetype=FileType.STREAM)
    it = iter(reader)

    (comm, path) = next(it)
    assert comm.id == 'one'

    (comm, path) = next(it)
    assert comm.id == 'two'

    (comm, path) = next(it)
    assert comm.id == 'three'

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False

    p.join()


@mark.posix
def test_concatenated_bz2(fifo):
    '''
    Note: concatenated_gz does not work, complaining about a tell (seek).
    tar_gz does work because the r|gz mode in tarfile results in direct
    calls to zlib for decompression.  gzip (which wraps zlib and is used
    in CommunicationReader for non-tar gz files) is the culprit.
    '''

    input_path = 'tests/testdata/simple_concatenated.bz2'
    p = Process(target=write_fifo, args=(input_path, fifo))
    p.start()

    reader = CommunicationReader(fifo, filetype=FileType.STREAM_BZ2)
    it = iter(reader)

    (comm, path) = next(it)
    assert comm.id == 'one'

    (comm, path) = next(it)
    assert comm.id == 'two'

    (comm, path) = next(it)
    assert comm.id == 'three'

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False

    p.join()


@mark.posix
def test_tar(fifo):
    input_path = 'tests/testdata/simple.tar'
    p = Process(target=write_fifo, args=(input_path, fifo))
    p.start()

    reader = CommunicationReader(fifo, filetype=FileType.TAR)
    it = iter(reader)

    (comm, path) = next(it)
    assert comm.id == 'one'

    (comm, path) = next(it)
    assert comm.id == 'two'

    (comm, path) = next(it)
    assert comm.id == 'three'

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False

    p.join()


@mark.posix
def test_tar_gz(fifo):
    input_path = 'tests/testdata/simple.tar.gz'
    p = Process(target=write_fifo, args=(input_path, fifo))
    p.start()

    reader = CommunicationReader(fifo, filetype=FileType.TAR_GZ)
    it = iter(reader)

    (comm, path) = next(it)
    assert comm.id == 'one'

    (comm, path) = next(it)
    assert comm.id == 'two'

    (comm, path) = next(it)
    assert comm.id == 'three'

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False

    p.join()


@mark.posix
def test_tar_bz2(fifo):
    input_path = 'tests/testdata/simple.tar.bz2'
    p = Process(target=write_fifo, args=(input_path, fifo))
    p.start()

    reader = CommunicationReader(fifo, filetype=FileType.TAR_BZ2)
    it = iter(reader)

    (comm, path) = next(it)
    assert comm.id == 'one'

    (comm, path) = next(it)
    assert comm.id == 'two'

    (comm, path) = next(it)
    assert comm.id == 'three'

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False

    p.join()



"""Add reference variables for each :class:`.UUID` "pointer" in a
:class:`.Communication`

"""
from __future__ import unicode_literals


def add_references_to_communication(comm):
    """Create references for each :class:`.UUID` 'pointer'

    Args:
        comm (Communication): A Concrete Communication object,
            will be modified by this function

    The Concrete schema uses :class:`.UUID` objects as internal
    pointers between Concrete objects.  This function adds member
    variables to Concrete objects that are references to the Concrete
    objects identified by the :class:`.UUID`.

    For example, each :class:`.Entity` has a `mentionIdlist` that
    lists the UUIDs of the :class:`.EntityMention` objects for that
    :class:`.Entity`.  This function adds a `mentionList` variable
    to the :class:`.Entity` that is a list of references to the
    actual :class:`.EntityMention` objects.  This allows you to
    access the :class:`.EntityMention` objects using:

        entity.mentionList

    This function adds these reference variables:

    - `tokenization` to each :class:`.TokenRefSequence`
    - `entityMention` to each :class:`.Argument`
    - `sentence` backpointer to each :class:`.Tokenization`
    - `parentMention` backpointer to appropriate :class:`.EntityMention`

    And adds these lists of reference variables:

    - `mentionList` to each :class:`.Entity`
    - `situationMention` to each :class:`.Argument`
    - `mentionList` to each :class:`.Situation`
    - `childMentionList` to each :class:`.EntityMention`

    For variables that represent optional lists of :class:`.UUID`
    objects (e.g. `situation.mentionIdList`), Python Thrift will set
    the variable to `None` if the list is not provided.  When this
    function adds a list-of-references variable (in this case,
    `situation.mentionList`) for an *omitted* optional list, it sets
    the new variable to `None` - it **DOES NOT** leave the variable
    undefined.

    """
#    comm.dependencyParseForUUID = {}
    comm.entityForUUID = {}
    comm.entityMentionForUUID = {}
#    comm.parseForUUID = {}
    comm.sectionForUUID = {}
    comm.sentenceForUUID = {}
    comm.situationForUUID = {}
    comm.situationMentionForUUID = {}
    comm.tokenizationForUUID = {}
#    comm.tokenTaggingForUUID = {}

    if comm.sectionList:
        for section in comm.sectionList:
            comm.sectionForUUID[section.uuid.uuidString] = section
            if section.sentenceList:
                for sentence in section.sentenceList:
                    comm.sentenceForUUID[sentence.uuid.uuidString] = sentence
                    if sentence.tokenization:
                        comm.tokenizationForUUID[
                            sentence.tokenization.uuid.uuidString] = \
                            sentence.tokenization
                        sentence.tokenization.sentence = sentence

    if comm.entityMentionSetList:
        for entityMentionSet in comm.entityMentionSetList:
            for entityMention in entityMentionSet.mentionList:
                comm.entityMentionForUUID[entityMention.uuid.uuidString] = \
                    entityMention
                try:
                    entityMention.tokens.tokenization = \
                        comm.tokenizationForUUID[
                            entityMention.tokens.tokenizationId.uuidString]
                except KeyError:
                    entityMention.tokens.tokenization = None
                # childMentionList and parentMention are in-memory references,
                # and not part of the Concrete schema
                entityMention.childMentionList = []
                entityMention.parentMention = None
                entityMention.entityMentionSet = entityMentionSet
            for entityMention in entityMentionSet.mentionList:
                if entityMention.childMentionIdList:
                    for childMentionId in entityMention.childMentionIdList:
                        childMention = comm.entityMentionForUUID[
                            childMentionId.uuidString]
                        childMention.parentMention = entityMention
                        entityMention.childMentionList.append(childMention)

    if comm.entitySetList:
        for entitySet in comm.entitySetList:
            for entity in entitySet.entityList:
                comm.entityForUUID[entity.uuid.uuidString] = entity
                entity.mentionList = []
                for mentionId in entity.mentionIdList:
                    entity.mentionList.append(
                        comm.entityMentionForUUID[mentionId.uuidString])
                entity.entitySet = entitySet

    if comm.situationMentionSetList:
        for situationMentionSet in comm.situationMentionSetList:
            for situationMention in situationMentionSet.mentionList:
                comm.situationMentionForUUID[situationMention.uuid.uuidString]\
                    = situationMention
                for argument in situationMention.argumentList:
                    if argument.entityMentionId:
                        argument.entityMention = comm.entityMentionForUUID[
                            argument.entityMentionId.uuidString]
                    else:
                        argument.entityMention = None
                    if argument.situationMentionId:
                        argument.situationMention = \
                            comm.situationMentionForUUID[
                                argument.situationMentionId.uuidString]
                    else:
                        argument.situationMention = None
                    if argument.tokens:
                        argument.tokens.tokenization = \
                            comm.tokenizationForUUID[
                                argument.tokens.tokenizationId.uuidString]
                if situationMention.tokens:
                    try:
                        situationMention.tokens.tokenization = \
                            comm.tokenizationForUUID[
                                situationMention.tokens.
                                tokenizationId.uuidString
                            ]
                    except KeyError:
                        situationMention.tokens.tokenization = None
                situationMention.situationMentionSet = situationMentionSet

    if comm.situationSetList:
        for situationSet in comm.situationSetList:
            for situation in situationSet.situationList:
                comm.situationForUUID[situation.uuid.uuidString] = situation
                if situation.mentionIdList:
                    situation.mentionList = []
                    for mentionId in situation.mentionIdList:
                        situation.mentionList.append(
                            comm.situationMentionForUUID[mentionId.uuidString])
                else:
                    situation.mentionList = None
                situation.situationSet = situationSet



from __future__ import unicode_literals
from concrete.util import (
    CommunicationContainerFetchHandler,
    RelayFetchHandler)
from concrete.util import (
    FetchCommunicationClientWrapper,
    SubprocessFetchCommunicationServiceWrapper)
from concrete.util import FetchBackedCommunicationContainer
from concrete.util import find_port
from concrete.util import create_comm
from concrete.validate import validate_communication


def test_comm_container_fetch_handler():
    comm_container = {
        'one': create_comm('one'),
        'two': create_comm('two')
    }

    impl = CommunicationContainerFetchHandler(comm_container)
    host = 'localhost'
    port = find_port()

    with SubprocessFetchCommunicationServiceWrapper(impl, host, port):
        with FetchCommunicationClientWrapper(host, port) as cli:
            assert cli.getCommunicationCount()
            ids = cli.getCommunicationIDs(0, 10)
            assert 'one' in ids
            assert 'two' in ids
            assert 'foo' not in ids


def test_fetch_backed_container():
    comm_container = {
        'one': create_comm('one'),
        'two': create_comm('two')
    }

    impl = CommunicationContainerFetchHandler(comm_container)
    host = 'localhost'
    port = find_port()

    with SubprocessFetchCommunicationServiceWrapper(impl, host, port):
        cc = FetchBackedCommunicationContainer(host, port)
        assert len(cc) == 2
        assert 'one' in cc
        assert 'two' in cc
        for comm_id in cc:
            comm = cc[comm_id]
            assert validate_communication(comm)


def test_relay_container_fetch_handler():
    comm_container = {
        'one': create_comm('one'),
        'two': create_comm('two')
    }

    impl = CommunicationContainerFetchHandler(comm_container)
    host = 'localhost'
    port = find_port()

    with SubprocessFetchCommunicationServiceWrapper(impl, host, port):
        relay_impl = RelayFetchHandler(host, port)
        relay_host = 'localhost'
        relay_port = find_port()

        with SubprocessFetchCommunicationServiceWrapper(relay_impl, relay_host, relay_port):
            with FetchCommunicationClientWrapper(relay_host, relay_port) as cli:
                assert cli.getCommunicationCount()
                ids = cli.getCommunicationIDs(0, 10)
                assert 'one' in ids
                assert 'two' in ids
                assert 'foo' not in ids



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.services.Service
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(concrete.services.Service.Iface):
    """
    The active learner client implements a method to accept new sorts of the annotation units
    """
    def submitSort(self, sessionId, unitIds):
        """
        Submit a new sort of communications to the broker

        Parameters:
         - sessionId
         - unitIds
        """
        pass


class Client(concrete.services.Service.Client, Iface):
    """
    The active learner client implements a method to accept new sorts of the annotation units
    """
    def __init__(self, iprot, oprot=None):
        concrete.services.Service.Client.__init__(self, iprot, oprot)

    def submitSort(self, sessionId, unitIds):
        """
        Submit a new sort of communications to the broker

        Parameters:
         - sessionId
         - unitIds
        """
        self.send_submitSort(sessionId, unitIds)
        self.recv_submitSort()

    def send_submitSort(self, sessionId, unitIds):
        self._oprot.writeMessageBegin('submitSort', TMessageType.CALL, self._seqid)
        args = submitSort_args()
        args.sessionId = sessionId
        args.unitIds = unitIds
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_submitSort(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = submitSort_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return


class Processor(concrete.services.Service.Processor, Iface, TProcessor):
    def __init__(self, handler):
        concrete.services.Service.Processor.__init__(self, handler)
        self._processMap["submitSort"] = Processor.process_submitSort

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_submitSort(self, seqid, iprot, oprot):
        args = submitSort_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = submitSort_result()
        try:
            self._handler.submitSort(args.sessionId, args.unitIds)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("submitSort", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class submitSort_args(object):
    """
    Attributes:
     - sessionId
     - unitIds
    """


    def __init__(self, sessionId=None, unitIds=None,):
        self.sessionId = sessionId
        self.unitIds = unitIds

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.sessionId = concrete.uuid.ttypes.UUID()
                    self.sessionId.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.unitIds = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = concrete.services.ttypes.AnnotationUnitIdentifier()
                        _elem19.read(iprot)
                        self.unitIds.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('submitSort_args')
        if self.sessionId is not None:
            oprot.writeFieldBegin('sessionId', TType.STRUCT, 1)
            self.sessionId.write(oprot)
            oprot.writeFieldEnd()
        if self.unitIds is not None:
            oprot.writeFieldBegin('unitIds', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.unitIds))
            for iter20 in self.unitIds:
                iter20.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(submitSort_args)
submitSort_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'sessionId', [concrete.uuid.ttypes.UUID, None], None, ),  # 1
    (2, TType.LIST, 'unitIds', (TType.STRUCT, [concrete.services.ttypes.AnnotationUnitIdentifier, None], False), None, ),  # 2
)


class submitSort_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('submitSort_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(submitSort_result)
submitSort_result.thrift_spec = (
)
fix_spec(all_structs)
del all_structs




#!/usr/bin/env python

"""
Command line script to (partially) validate a Concrete Communication

This script is a thin wrapper around the functionality in the
concrete.validate library.
"""
from __future__ import unicode_literals

import argparse
import logging

import networkx as nx

from concrete.util.tokenization import get_comm_tokenizations
import concrete.version
from concrete.util.file_io import CommunicationReader, FileType
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Check dependency parse graph for errors")
    parser.add_argument('communication_file',
                        nargs='?',
                        type=str,
                        help='file to read communications from '
                             '(if not specified, read from standard input)')
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    # Won't work on Windows
    if args.communication_file is None:
        reader_kwargs = dict(filetype=FileType.STREAM)
        input_path = '/dev/fd/0'
    else:
        reader_kwargs = dict()
        input_path = args.communication_file

    for (comm, filename) in CommunicationReader(input_path, **reader_kwargs):
        logging.info(u"Inspecting Communication with ID '%s" % comm.id)
        for tokenization in get_comm_tokenizations(comm):
            inspect_dependency_parses(tokenization)


def inspect_dependency_parses(tokenization):

    def _get_token_text(tokenization):
        return " ".join([t.text for t in tokenization.tokenList.tokenList])

    if tokenization.dependencyParseList:
        total_tokens = len(tokenization.tokenList.tokenList)
        for dependencyParse in tokenization.dependencyParseList:
            dependency_parse_tree = nx.DiGraph()

            # Add nodes to dependency parse tree
            for dependency in dependencyParse.dependencyList:
                if (dependency.gov is None and
                        dependency.edgeType.lower() != "root"):
                    logging.error((u"  Found a null dependency parse node with"
                                   u" governer whose edgeType is '%s' instead"
                                   u" of 'root'") % dependency.edgeType)
                if dependency.gov is not None:
                    if dependency.gov < -1 or dependency.gov > total_tokens:
                        logging.error(u"  Found a null dependency parse node"
                                      u" with invalid governer of '%d'" %
                                      dependency.gov)
                    dependency_parse_tree.add_node(dependency.gov)
                dependency_parse_tree.add_node(dependency.dep)

            # Add edges to dependency parse tree
            for dependency in dependencyParse.dependencyList:
                if dependency.gov is not None:
                    dependency_parse_tree.add_edge(
                        dependency.gov, dependency.dep)

            # Check if dependency parse "tree" is a fully connected graph
            undirected_graph = dependency_parse_tree.to_undirected()
            try:
                if not nx.is_connected(undirected_graph):
                    logging.error((u"  The dependency parse graph created by"
                                   u" '%s' is not a fully connected graph -"
                                   u" the graph has %d components. Token text:"
                                   u" '%s'") %
                                  (dependencyParse.metadata.tool,
                                   nx.number_connected_components(
                                       undirected_graph),
                                   _get_token_text(tokenization)))
            except nx.exception.NetworkXPointlessConcept:
                logging.warning((u"  The dependency parse graph created by"
                                 u" '%s' does not have any nodes. Token text:"
                                 u" '%s'") %
                                (dependencyParse.metadata.tool,
                                 _get_token_text(tokenization)))


if __name__ == "__main__":
    main()



"""Convert Concrete objects to JSON strings
"""
from __future__ import unicode_literals

import json

from thrift import TSerialization
from thrift.protocol import TJSONProtocol

from .file_io import (
    read_communication_from_file,
    read_tokenlattice_from_file,
)


def communication_file_to_json(communication_filename, remove_timestamps=False,
                               remove_uuids=False):
    """Get a "pretty-printed" JSON string representation for a :class:`.Communication`

    Args:
        communication_filename (str): Communication filename
        remove_timestamps (bool): Flag for removing timestamps from JSON output
        remove_uuids (bool): Flag for removing :class:`.UUID` info from JSON output

    Returns:
        str: A "pretty-printed" JSON representation of the Communication
    """
    comm = read_communication_from_file(communication_filename)
    return thrift_to_json(comm, remove_timestamps=remove_timestamps,
                          remove_uuids=remove_uuids)


def tokenlattice_file_to_json(toklat_filename):
    """Get a "pretty-printed" JSON string representation for a :class:`.TokenLattice`

    Args:
        toklat_filename (str): String specifying TokenLattice filename

    Returns:
        str: A "pretty-printed" JSON representation of the TokenLattice
    """
    toklat = read_tokenlattice_from_file(toklat_filename)
    return thrift_to_json(toklat)


def get_json_object_without_timestamps(json_object):
    """Create a copy of a JSON object created by `json.loads()`, with all
    representations of :class:`.AnnotationMetadata` timestamps
    (dictionary keys with value `timestamp`) recursively removed.

    Args:
        json_object: Python object created from string by `json.loads()`

    Returns:
         A copy of the input data structure with all timestamp objects removed
    """
    if type(json_object) is list:
        new_json_object = [get_json_object_without_timestamps(obj)
                           for obj in json_object]
    elif type(json_object) is dict:
        new_json_object = {}
        for k, v in json_object.items():
            if k != "timestamp":
                new_json_object[k] = get_json_object_without_timestamps(v)
    else:
        new_json_object = json_object

    return new_json_object


def get_json_object_without_uuids(json_object):
    """Create a copy of a JSON object created by `json.loads()`, with all
    representations of :class:`.UUID` objects (dictionaries containing a
    'uuidString' key) recursively removed.

    Args:
        json_object: Python object created from string by `json.loads()`

    Returns:
        A copy of the input data structure with all UUID objects removed
    """
    if type(json_object) is list:
        new_json_object = [get_json_object_without_uuids(obj)
                           for obj in json_object
                           if type(obj) is not dict or 'uuidString' not in obj]
    elif type(json_object) is dict:
        new_json_object = {}
        for k, v in json_object.items():
            if type(v) is not dict or 'uuidString' not in v:
                new_json_object[k] = get_json_object_without_uuids(v)
    else:
        new_json_object = json_object

    return new_json_object


def thrift_to_json(tobj, remove_timestamps=False, remove_uuids=False):
    """Get a "pretty-printed" JSON string representation for a Thrift object

    Args:
        tobj: A Thrift object
        remove_timestamps (bool): Flag for removing timestamps from JSON output
        remove_uuids (bool): Flag for removing :class:`.UUID` info from JSON output

    Returns:
        str: A "pretty-printed" JSON representation of the Thrift object
    """
    thrift_json_string = TSerialization.serialize(
        tobj, TJSONProtocol.TSimpleJSONProtocolFactory()).decode('utf-8')
    thrift_json = json.loads(thrift_json_string)

    if remove_timestamps:
        thrift_json = get_json_object_without_timestamps(thrift_json)
    if remove_uuids:
        thrift_json = get_json_object_without_uuids(thrift_json)

    return json.dumps(thrift_json, indent=2, separators=(',', ': '),
                      ensure_ascii=False, sort_keys=True)



#!/usr/bin/env python

'Pretty-prints a Concrete file as JSON'
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import codecs
import logging

from thrift import TSerialization
from thrift.protocol import TJSONProtocol

from concrete.util import (
    communication_file_to_json,
    tokenlattice_file_to_json,
    read_communication_from_file)
from concrete.util import set_stdout_encoding
import concrete.version


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Pretty Print a Concrete file")
    parser.add_argument('--concrete_type', default='communication',
                        choices=['communication', 'tokenlattice'],
                        help='Default: communication')
    parser.add_argument('--protocol', default='simple',
                        choices=['simple', 'TJSONProtocol'],
                        help='Default: simple')
    parser.add_argument('--remove-timestamps', action='store_true',
                        help="Removes timestamps from JSON output")
    parser.add_argument('--remove-uuids', action='store_true',
                        help="Removes UUIDs from JSON output")
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    parser.add_argument('concrete_file',
                        help='path to input concrete communication file')
    parser.add_argument('json_file', nargs='?', default='-',
                        help='path to output json file')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    if args.protocol == 'simple':
        if args.concrete_type == 'communication':
            json_communication = communication_file_to_json(
                args.concrete_file,
                remove_timestamps=args.remove_timestamps,
                remove_uuids=args.remove_uuids
            )
        else:
            json_communication = tokenlattice_file_to_json(
                args.concrete_file
            )
    else:
        if args.concrete_type == 'communication':
            comm = read_communication_from_file(args.concrete_file)
            json_communication = TSerialization.serialize(
                comm, TJSONProtocol.TJSONProtocolFactory()).decode('utf-8')
        else:
            raise NotImplementedError

    if args.json_file == '-':
        print(json_communication)
    else:
        with codecs.open(args.json_file, 'w', encoding='utf-8') as f:
            f.write(json_communication)


if __name__ == "__main__":
    main()



from ..learn import (
    ActiveLearnerClientService,
    ActiveLearnerServerService)
from .service_wrapper import (
    ConcreteServiceClientWrapper,
    ConcreteServiceWrapper,
    HTTPConcreteServiceClientWrapper,
    SubprocessConcreteServiceWrapper)


ActiveLearnerClientClientWrapper = type(
    'ActiveLearnerClientClientWrapper',
    (ConcreteServiceClientWrapper,),
    {'concrete_service_class': ActiveLearnerClientService})


ActiveLearnerClientServiceWrapper = type(
    'ActiveLearnerClientServiceWrapper',
    (ConcreteServiceWrapper,),
    {'concrete_service_class': ActiveLearnerClientService})


HTTPActiveLearnerClientClientWrapper = type(
    'HTTPActiveLearnerClientClientWrapper',
    (HTTPConcreteServiceClientWrapper,),
    {'concrete_service_class': ActiveLearnerClientService})


SubprocessActiveLearnerClientServiceWrapper = type(
    'SubprocessActiveLearnerClientServiceWrapper',
    (SubprocessConcreteServiceWrapper,),
    {'concrete_service_wrapper_class': ActiveLearnerClientServiceWrapper})


ActiveLearnerServerClientWrapper = type(
    'ActiveLearnerServerClientWrapper',
    (ConcreteServiceClientWrapper,),
    {'concrete_service_class': ActiveLearnerServerService})


ActiveLearnerServerServiceWrapper = type(
    'ActiveLearnerServerServiceWrapper',
    (ConcreteServiceWrapper,),
    {'concrete_service_class': ActiveLearnerServerService})


HTTPActiveLearnerServerClientWrapper = type(
    'HTTPActiveLearnerServerClientWrapper',
    (HTTPConcreteServiceClientWrapper,),
    {'concrete_service_class': ActiveLearnerServerService})


SubprocessActiveLearnerServerServiceWrapper = type(
    'SubprocessActiveLearnerServerServiceWrapper',
    (SubprocessConcreteServiceWrapper,),
    {'concrete_service_wrapper_class': ActiveLearnerServerServiceWrapper})



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class NITFInfo(object):
    """
    Attributes:
     - alternateURL: This field specifies the URL of the article, if published online. In some
    cases, such as with the New York Times, when this field is present,
    the URL is preferred to the URL field on articles published on
    or after April 02, 2006, as the linked page will have richer content.
     - articleAbstract: This field is a summary of the article, possibly written by
    an indexing service.
     - authorBiography: This field specifies the biography of the author of the article.
    Generally, this field is specified for guest authors, and not for
    regular reporters, except to provide the author's email address.
     - banner: The banner field is used to indicate if there has been additional
    information appended to the articles since its publication. Examples of
    banners include ('Correction Appended' and 'Editor's Note Appended').
     - biographicalCategoryList: When present, the biographical category field generally indicates that a
    document focuses on a particular individual. The value of the field
    indicates the area or category in which this individual is best known.
    This field is most often defined for Obituaries and Book Reviews.

    <ol>
    <li>Politics and Government (U.S.)</li>
    <li>Books and Magazines <li>Royalty</li>
    </ol>
     - columnName: If the article is part of a regular column, this field specifies the name
    of that column.
    <br>
    Sample Column Names:
    <br>
    <ol>
    <li>World News Briefs</li>
    <li>WEDDINGS</li>
    <li>The Accessories Channel</li>
    </ol>

     - columnNumber: This field specifies the column in which the article starts in the print
    paper. A typical printed page in the paper has six columns numbered from
    right to left. As a consequence most, but not all, of the values for this
    field fall in the range 1-6.
     - correctionDate: This field specifies the date on which a correction was made to the
    article. Generally, if the correction date is specified, the correction
    text will also be specified (and vice versa).
     - correctionText: For articles corrected following publication, this field specifies the
    correction. Generally, if the correction text is specified, the
    correction date will also be specified (and vice versa).
     - credit: This field indicates the entity that produced the editorial content of
    this document.
     - dayOfWeek: This field specifies the day of week on which the article was published.
    <ul>
    <li>Monday</li>
    <li>Tuesday</li>
    <li>Wednesday</li>
    <li>Thursday</li>
    <li>Friday</li>
    <li>Saturday</li>
    <li>Sunday</li>
    </ul>
     - descriptorList: The &quot;descriptors&quot; field specifies a list of descriptive terms drawn from
    a normalized controlled vocabulary corresponding to subjects mentioned in
    the article.
    <br>
    Examples Include:
    <ol>
    <li>ECONOMIC CONDITIONS AND TRENDS</li>
    <li>AIRPLANES</li>
    <li>VIOLINS</li>
    </ol>
     - featurePage: The feature page containing this article, such as
    <ul>
    <li>Education Page</li>
    <li>Fashion Page</li>
    </ul>
     - generalOnlineDescriptorList: The &quot;general online descriptors&quot; field specifies a list of descriptors
    that are at a higher level of generality than the other tags associated
    with the article.
    <br>
    Examples Include:
    <ol>
    <li>Surfing</li>
    <li>Venice Biennale</li>
    <li>Ranches</li>
    </ol>
     - guid: The GUID field specifies an integer that is guaranteed to be unique for
    every document in the corpus.
     - kicker: The kicker is an additional piece of information printed as an
    accompaniment to a news headline.
     - leadParagraphList: The &quot;lead Paragraph&quot; field is the lead paragraph of the article.
    Generally this field is populated with the first two paragraphs from the
    article.
     - locationList: The &quot;locations&quot; field specifies a list of geographic descriptors drawn
    from a normalized controlled vocabulary that correspond to places
    mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Wellsboro (Pa)</li>
    <li>Kansas City (Kan)</li>
    <li>Park Slope (NYC)</li>
    </ol>
     - nameList: The &quot;names&quot; field specifies a list of names mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Azza Fahmy</li>
    <li>George C. Izenour</li>
    <li>Chris Schenkel</li>
    </ol>
     - newsDesk: This field specifies the desk in the newsroom that
    produced the article. The desk is related to, but is not the same as the
    section in which the article appears.
     - normalizedByline: The Normalized Byline field is the byline normalized to the form (last
    name, first name).
     - onlineDescriptorList: This field specifies a list of descriptors from a normalized controlled
    vocabulary that correspond to topics mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Marriages</li>
    <li>Parks and Other Recreation Areas</li>
    <li>Cooking and Cookbooks</li>
    </ol>
     - onlineHeadline: This field specifies the headline displayed with the article
    online. Often this differs from the headline used in print.
     - onlineLeadParagraph: This field specifies the lead paragraph for the online version.
     - onlineLocationList: This field specifies a list of place names that correspond to geographic
    locations mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Hollywood</li>
    <li>Los Angeles</li>
    <li>Arcadia</li>
    </ol>
     - onlineOrganizationList: This field specifies a list of organizations that correspond to
    organizations mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Nintendo Company Limited</li>
    <li>Yeshiva University</li>
    <li>Rose Center</li>
    </ol>
     - onlinePeople: This field specifies a list of people that correspond to individuals
    mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Lopez, Jennifer</li>
    <li>Joyce, James</li>
    <li>Robinson, Jackie</li>
    </ol>
     - onlineSectionList: This field specifies the section(s) in which the online version of the article
    is placed. This may typically be populated from a semicolon (;) delineated list.
     - onlineTitleList: This field specifies a list of authored works mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Matchstick Men (Movie)</li>
    <li>Blades of Glory (Movie)</li>
    <li>Bridge and Tunnel (Play)</li>
    </ol>
     - organizationList: This field specifies a list of organization names drawn from a normalized
    controlled vocabulary that correspond to organizations mentioned in the
    article.
    <br>
    Examples Include:
    <ol>
    <li>Circuit City Stores Inc</li>
    <li>Delaware County Community College (Pa)</li>
    <li>CONNECTICUT GRAND OPERA</li>
    </ol>
     - page: This field specifies the page of the section in the paper in which the
    article appears. This is not an absolute pagination. An article that
    appears on page 3 in section A occurs in the physical paper before an
    article that occurs on page 1 of section F. The section is encoded in
    the <strong>section</strong> field.
     - peopleList: This field specifies a list of people from a normalized controlled
    vocabulary that correspond to individuals mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>REAGAN, RONALD WILSON (PRES)</li>
    <li>BEGIN, MENACHEM (PRIME MIN)</li>
    <li>COLLINS, GLENN</li>
    </ol>
     - publicationDate: This field specifies the date of the article's publication.
     - publicationDayOfMonth: This field specifies the day of the month on which the article was
    published, always in the range 1-31.
     - publicationMonth: This field specifies the month on which the article was published in the
    range 1-12 where 1 is January 2 is February etc.
     - publicationYear: This field specifies the year in which the article was published. This
    value is in the range 1987-2007 for this collection.
     - section: This field specifies the section of the paper in which the article
    appears. This is not the name of the section, but rather a letter or
    number that indicates the section.
     - seriesName: If the article is part of a regular series, this field specifies the name
    of that column.
     - slug: The slug is a short string that uniquely identifies an article from all
    other articles published on the same day. Please note, however, that
    different articles on different days may have the same slug.
    <ul>
    <li>30other</li>
    <li>12reunion</li>
    </ul>
     - taxonomicClassifierList: This field specifies a list of taxonomic classifiers that place this
    article into a hierarchy of articles. The individual terms of each
    taxonomic classifier are separated with the '/' character.
    <br>
    Examples Include:
    <ol>
    <li>Top/Features/Travel/Guides/Destinations/North America/United
    States/Arizona</li>
    <li>Top/News/U.S./Rockies</li>
    <li>Top/Opinion</li>
    </ol>
     - titleList: This field specifies a list of authored works that correspond to works
    mentioned in the article.
    <br>
    Examples Include:
    <ol>
    <li>Greystoke: The Legend of Tarzan, Lord of the Apes (Movie)</li>
    <li>Law and Order (TV Program)</li>
    <li>BATTLEFIELD EARTH (BOOK)</li>
    </ol>
     - typesOfMaterialList: This field specifies a normalized list of terms describing the general
    editorial category of the article.
    <br>
    Examples Include:
    <ol>
    <li>REVIEW</li>
    <li>OBITUARY</li>
    <li>ANALYSIS</li>
    </ol>
     - url: This field specifies the location of the online version of the article. The
    &quot;Alternative Url&quot; field is preferred to this field on articles published
    on or after April 02, 2006, as the linked page will have richer content.
     - wordCount: This field specifies the number of words in the body of the article,
    including the lead paragraph.
    """


    def __init__(self, alternateURL=None, articleAbstract=None, authorBiography=None, banner=None, biographicalCategoryList=None, columnName=None, columnNumber=None, correctionDate=None, correctionText=None, credit=None, dayOfWeek=None, descriptorList=None, featurePage=None, generalOnlineDescriptorList=None, guid=None, kicker=None, leadParagraphList=None, locationList=None, nameList=None, newsDesk=None, normalizedByline=None, onlineDescriptorList=None, onlineHeadline=None, onlineLeadParagraph=None, onlineLocationList=None, onlineOrganizationList=None, onlinePeople=None, onlineSectionList=None, onlineTitleList=None, organizationList=None, page=None, peopleList=None, publicationDate=None, publicationDayOfMonth=None, publicationMonth=None, publicationYear=None, section=None, seriesName=None, slug=None, taxonomicClassifierList=None, titleList=None, typesOfMaterialList=None, url=None, wordCount=None,):
        self.alternateURL = alternateURL
        self.articleAbstract = articleAbstract
        self.authorBiography = authorBiography
        self.banner = banner
        self.biographicalCategoryList = biographicalCategoryList
        self.columnName = columnName
        self.columnNumber = columnNumber
        self.correctionDate = correctionDate
        self.correctionText = correctionText
        self.credit = credit
        self.dayOfWeek = dayOfWeek
        self.descriptorList = descriptorList
        self.featurePage = featurePage
        self.generalOnlineDescriptorList = generalOnlineDescriptorList
        self.guid = guid
        self.kicker = kicker
        self.leadParagraphList = leadParagraphList
        self.locationList = locationList
        self.nameList = nameList
        self.newsDesk = newsDesk
        self.normalizedByline = normalizedByline
        self.onlineDescriptorList = onlineDescriptorList
        self.onlineHeadline = onlineHeadline
        self.onlineLeadParagraph = onlineLeadParagraph
        self.onlineLocationList = onlineLocationList
        self.onlineOrganizationList = onlineOrganizationList
        self.onlinePeople = onlinePeople
        self.onlineSectionList = onlineSectionList
        self.onlineTitleList = onlineTitleList
        self.organizationList = organizationList
        self.page = page
        self.peopleList = peopleList
        self.publicationDate = publicationDate
        self.publicationDayOfMonth = publicationDayOfMonth
        self.publicationMonth = publicationMonth
        self.publicationYear = publicationYear
        self.section = section
        self.seriesName = seriesName
        self.slug = slug
        self.taxonomicClassifierList = taxonomicClassifierList
        self.titleList = titleList
        self.typesOfMaterialList = typesOfMaterialList
        self.url = url
        self.wordCount = wordCount

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.alternateURL = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.articleAbstract = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.authorBiography = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.banner = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.LIST:
                    self.biographicalCategoryList = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.biographicalCategoryList.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.columnName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.I32:
                    self.columnNumber = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.I64:
                    self.correctionDate = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.STRING:
                    self.correctionText = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 10:
                if ftype == TType.STRING:
                    self.credit = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 11:
                if ftype == TType.STRING:
                    self.dayOfWeek = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 12:
                if ftype == TType.LIST:
                    self.descriptorList = []
                    (_etype9, _size6) = iprot.readListBegin()
                    for _i10 in range(_size6):
                        _elem11 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.descriptorList.append(_elem11)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 13:
                if ftype == TType.STRING:
                    self.featurePage = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 14:
                if ftype == TType.LIST:
                    self.generalOnlineDescriptorList = []
                    (_etype15, _size12) = iprot.readListBegin()
                    for _i16 in range(_size12):
                        _elem17 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.generalOnlineDescriptorList.append(_elem17)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 15:
                if ftype == TType.I32:
                    self.guid = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 16:
                if ftype == TType.STRING:
                    self.kicker = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 17:
                if ftype == TType.LIST:
                    self.leadParagraphList = []
                    (_etype21, _size18) = iprot.readListBegin()
                    for _i22 in range(_size18):
                        _elem23 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.leadParagraphList.append(_elem23)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 18:
                if ftype == TType.LIST:
                    self.locationList = []
                    (_etype27, _size24) = iprot.readListBegin()
                    for _i28 in range(_size24):
                        _elem29 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.locationList.append(_elem29)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 19:
                if ftype == TType.LIST:
                    self.nameList = []
                    (_etype33, _size30) = iprot.readListBegin()
                    for _i34 in range(_size30):
                        _elem35 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.nameList.append(_elem35)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 20:
                if ftype == TType.STRING:
                    self.newsDesk = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 21:
                if ftype == TType.STRING:
                    self.normalizedByline = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 22:
                if ftype == TType.LIST:
                    self.onlineDescriptorList = []
                    (_etype39, _size36) = iprot.readListBegin()
                    for _i40 in range(_size36):
                        _elem41 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlineDescriptorList.append(_elem41)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 23:
                if ftype == TType.STRING:
                    self.onlineHeadline = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 24:
                if ftype == TType.STRING:
                    self.onlineLeadParagraph = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 25:
                if ftype == TType.LIST:
                    self.onlineLocationList = []
                    (_etype45, _size42) = iprot.readListBegin()
                    for _i46 in range(_size42):
                        _elem47 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlineLocationList.append(_elem47)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 26:
                if ftype == TType.LIST:
                    self.onlineOrganizationList = []
                    (_etype51, _size48) = iprot.readListBegin()
                    for _i52 in range(_size48):
                        _elem53 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlineOrganizationList.append(_elem53)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 27:
                if ftype == TType.LIST:
                    self.onlinePeople = []
                    (_etype57, _size54) = iprot.readListBegin()
                    for _i58 in range(_size54):
                        _elem59 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlinePeople.append(_elem59)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 28:
                if ftype == TType.LIST:
                    self.onlineSectionList = []
                    (_etype63, _size60) = iprot.readListBegin()
                    for _i64 in range(_size60):
                        _elem65 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlineSectionList.append(_elem65)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 29:
                if ftype == TType.LIST:
                    self.onlineTitleList = []
                    (_etype69, _size66) = iprot.readListBegin()
                    for _i70 in range(_size66):
                        _elem71 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.onlineTitleList.append(_elem71)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 30:
                if ftype == TType.LIST:
                    self.organizationList = []
                    (_etype75, _size72) = iprot.readListBegin()
                    for _i76 in range(_size72):
                        _elem77 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.organizationList.append(_elem77)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 31:
                if ftype == TType.I32:
                    self.page = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 32:
                if ftype == TType.LIST:
                    self.peopleList = []
                    (_etype81, _size78) = iprot.readListBegin()
                    for _i82 in range(_size78):
                        _elem83 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.peopleList.append(_elem83)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 33:
                if ftype == TType.I64:
                    self.publicationDate = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 34:
                if ftype == TType.I32:
                    self.publicationDayOfMonth = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 35:
                if ftype == TType.I32:
                    self.publicationMonth = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 36:
                if ftype == TType.I32:
                    self.publicationYear = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 37:
                if ftype == TType.STRING:
                    self.section = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 38:
                if ftype == TType.STRING:
                    self.seriesName = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 39:
                if ftype == TType.STRING:
                    self.slug = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 40:
                if ftype == TType.LIST:
                    self.taxonomicClassifierList = []
                    (_etype87, _size84) = iprot.readListBegin()
                    for _i88 in range(_size84):
                        _elem89 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.taxonomicClassifierList.append(_elem89)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 41:
                if ftype == TType.LIST:
                    self.titleList = []
                    (_etype93, _size90) = iprot.readListBegin()
                    for _i94 in range(_size90):
                        _elem95 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.titleList.append(_elem95)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 42:
                if ftype == TType.LIST:
                    self.typesOfMaterialList = []
                    (_etype99, _size96) = iprot.readListBegin()
                    for _i100 in range(_size96):
                        _elem101 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.typesOfMaterialList.append(_elem101)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 43:
                if ftype == TType.STRING:
                    self.url = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 44:
                if ftype == TType.I32:
                    self.wordCount = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('NITFInfo')
        if self.alternateURL is not None:
            oprot.writeFieldBegin('alternateURL', TType.STRING, 1)
            oprot.writeString(self.alternateURL.encode('utf-8') if sys.version_info[0] == 2 else self.alternateURL)
            oprot.writeFieldEnd()
        if self.articleAbstract is not None:
            oprot.writeFieldBegin('articleAbstract', TType.STRING, 2)
            oprot.writeString(self.articleAbstract.encode('utf-8') if sys.version_info[0] == 2 else self.articleAbstract)
            oprot.writeFieldEnd()
        if self.authorBiography is not None:
            oprot.writeFieldBegin('authorBiography', TType.STRING, 3)
            oprot.writeString(self.authorBiography.encode('utf-8') if sys.version_info[0] == 2 else self.authorBiography)
            oprot.writeFieldEnd()
        if self.banner is not None:
            oprot.writeFieldBegin('banner', TType.STRING, 4)
            oprot.writeString(self.banner.encode('utf-8') if sys.version_info[0] == 2 else self.banner)
            oprot.writeFieldEnd()
        if self.biographicalCategoryList is not None:
            oprot.writeFieldBegin('biographicalCategoryList', TType.LIST, 5)
            oprot.writeListBegin(TType.STRING, len(self.biographicalCategoryList))
            for iter102 in self.biographicalCategoryList:
                oprot.writeString(iter102.encode('utf-8') if sys.version_info[0] == 2 else iter102)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.columnName is not None:
            oprot.writeFieldBegin('columnName', TType.STRING, 6)
            oprot.writeString(self.columnName.encode('utf-8') if sys.version_info[0] == 2 else self.columnName)
            oprot.writeFieldEnd()
        if self.columnNumber is not None:
            oprot.writeFieldBegin('columnNumber', TType.I32, 7)
            oprot.writeI32(self.columnNumber)
            oprot.writeFieldEnd()
        if self.correctionDate is not None:
            oprot.writeFieldBegin('correctionDate', TType.I64, 8)
            oprot.writeI64(self.correctionDate)
            oprot.writeFieldEnd()
        if self.correctionText is not None:
            oprot.writeFieldBegin('correctionText', TType.STRING, 9)
            oprot.writeString(self.correctionText.encode('utf-8') if sys.version_info[0] == 2 else self.correctionText)
            oprot.writeFieldEnd()
        if self.credit is not None:
            oprot.writeFieldBegin('credit', TType.STRING, 10)
            oprot.writeString(self.credit.encode('utf-8') if sys.version_info[0] == 2 else self.credit)
            oprot.writeFieldEnd()
        if self.dayOfWeek is not None:
            oprot.writeFieldBegin('dayOfWeek', TType.STRING, 11)
            oprot.writeString(self.dayOfWeek.encode('utf-8') if sys.version_info[0] == 2 else self.dayOfWeek)
            oprot.writeFieldEnd()
        if self.descriptorList is not None:
            oprot.writeFieldBegin('descriptorList', TType.LIST, 12)
            oprot.writeListBegin(TType.STRING, len(self.descriptorList))
            for iter103 in self.descriptorList:
                oprot.writeString(iter103.encode('utf-8') if sys.version_info[0] == 2 else iter103)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.featurePage is not None:
            oprot.writeFieldBegin('featurePage', TType.STRING, 13)
            oprot.writeString(self.featurePage.encode('utf-8') if sys.version_info[0] == 2 else self.featurePage)
            oprot.writeFieldEnd()
        if self.generalOnlineDescriptorList is not None:
            oprot.writeFieldBegin('generalOnlineDescriptorList', TType.LIST, 14)
            oprot.writeListBegin(TType.STRING, len(self.generalOnlineDescriptorList))
            for iter104 in self.generalOnlineDescriptorList:
                oprot.writeString(iter104.encode('utf-8') if sys.version_info[0] == 2 else iter104)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.guid is not None:
            oprot.writeFieldBegin('guid', TType.I32, 15)
            oprot.writeI32(self.guid)
            oprot.writeFieldEnd()
        if self.kicker is not None:
            oprot.writeFieldBegin('kicker', TType.STRING, 16)
            oprot.writeString(self.kicker.encode('utf-8') if sys.version_info[0] == 2 else self.kicker)
            oprot.writeFieldEnd()
        if self.leadParagraphList is not None:
            oprot.writeFieldBegin('leadParagraphList', TType.LIST, 17)
            oprot.writeListBegin(TType.STRING, len(self.leadParagraphList))
            for iter105 in self.leadParagraphList:
                oprot.writeString(iter105.encode('utf-8') if sys.version_info[0] == 2 else iter105)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.locationList is not None:
            oprot.writeFieldBegin('locationList', TType.LIST, 18)
            oprot.writeListBegin(TType.STRING, len(self.locationList))
            for iter106 in self.locationList:
                oprot.writeString(iter106.encode('utf-8') if sys.version_info[0] == 2 else iter106)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.nameList is not None:
            oprot.writeFieldBegin('nameList', TType.LIST, 19)
            oprot.writeListBegin(TType.STRING, len(self.nameList))
            for iter107 in self.nameList:
                oprot.writeString(iter107.encode('utf-8') if sys.version_info[0] == 2 else iter107)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.newsDesk is not None:
            oprot.writeFieldBegin('newsDesk', TType.STRING, 20)
            oprot.writeString(self.newsDesk.encode('utf-8') if sys.version_info[0] == 2 else self.newsDesk)
            oprot.writeFieldEnd()
        if self.normalizedByline is not None:
            oprot.writeFieldBegin('normalizedByline', TType.STRING, 21)
            oprot.writeString(self.normalizedByline.encode('utf-8') if sys.version_info[0] == 2 else self.normalizedByline)
            oprot.writeFieldEnd()
        if self.onlineDescriptorList is not None:
            oprot.writeFieldBegin('onlineDescriptorList', TType.LIST, 22)
            oprot.writeListBegin(TType.STRING, len(self.onlineDescriptorList))
            for iter108 in self.onlineDescriptorList:
                oprot.writeString(iter108.encode('utf-8') if sys.version_info[0] == 2 else iter108)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.onlineHeadline is not None:
            oprot.writeFieldBegin('onlineHeadline', TType.STRING, 23)
            oprot.writeString(self.onlineHeadline.encode('utf-8') if sys.version_info[0] == 2 else self.onlineHeadline)
            oprot.writeFieldEnd()
        if self.onlineLeadParagraph is not None:
            oprot.writeFieldBegin('onlineLeadParagraph', TType.STRING, 24)
            oprot.writeString(self.onlineLeadParagraph.encode('utf-8') if sys.version_info[0] == 2 else self.onlineLeadParagraph)
            oprot.writeFieldEnd()
        if self.onlineLocationList is not None:
            oprot.writeFieldBegin('onlineLocationList', TType.LIST, 25)
            oprot.writeListBegin(TType.STRING, len(self.onlineLocationList))
            for iter109 in self.onlineLocationList:
                oprot.writeString(iter109.encode('utf-8') if sys.version_info[0] == 2 else iter109)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.onlineOrganizationList is not None:
            oprot.writeFieldBegin('onlineOrganizationList', TType.LIST, 26)
            oprot.writeListBegin(TType.STRING, len(self.onlineOrganizationList))
            for iter110 in self.onlineOrganizationList:
                oprot.writeString(iter110.encode('utf-8') if sys.version_info[0] == 2 else iter110)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.onlinePeople is not None:
            oprot.writeFieldBegin('onlinePeople', TType.LIST, 27)
            oprot.writeListBegin(TType.STRING, len(self.onlinePeople))
            for iter111 in self.onlinePeople:
                oprot.writeString(iter111.encode('utf-8') if sys.version_info[0] == 2 else iter111)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.onlineSectionList is not None:
            oprot.writeFieldBegin('onlineSectionList', TType.LIST, 28)
            oprot.writeListBegin(TType.STRING, len(self.onlineSectionList))
            for iter112 in self.onlineSectionList:
                oprot.writeString(iter112.encode('utf-8') if sys.version_info[0] == 2 else iter112)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.onlineTitleList is not None:
            oprot.writeFieldBegin('onlineTitleList', TType.LIST, 29)
            oprot.writeListBegin(TType.STRING, len(self.onlineTitleList))
            for iter113 in self.onlineTitleList:
                oprot.writeString(iter113.encode('utf-8') if sys.version_info[0] == 2 else iter113)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.organizationList is not None:
            oprot.writeFieldBegin('organizationList', TType.LIST, 30)
            oprot.writeListBegin(TType.STRING, len(self.organizationList))
            for iter114 in self.organizationList:
                oprot.writeString(iter114.encode('utf-8') if sys.version_info[0] == 2 else iter114)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.page is not None:
            oprot.writeFieldBegin('page', TType.I32, 31)
            oprot.writeI32(self.page)
            oprot.writeFieldEnd()
        if self.peopleList is not None:
            oprot.writeFieldBegin('peopleList', TType.LIST, 32)
            oprot.writeListBegin(TType.STRING, len(self.peopleList))
            for iter115 in self.peopleList:
                oprot.writeString(iter115.encode('utf-8') if sys.version_info[0] == 2 else iter115)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.publicationDate is not None:
            oprot.writeFieldBegin('publicationDate', TType.I64, 33)
            oprot.writeI64(self.publicationDate)
            oprot.writeFieldEnd()
        if self.publicationDayOfMonth is not None:
            oprot.writeFieldBegin('publicationDayOfMonth', TType.I32, 34)
            oprot.writeI32(self.publicationDayOfMonth)
            oprot.writeFieldEnd()
        if self.publicationMonth is not None:
            oprot.writeFieldBegin('publicationMonth', TType.I32, 35)
            oprot.writeI32(self.publicationMonth)
            oprot.writeFieldEnd()
        if self.publicationYear is not None:
            oprot.writeFieldBegin('publicationYear', TType.I32, 36)
            oprot.writeI32(self.publicationYear)
            oprot.writeFieldEnd()
        if self.section is not None:
            oprot.writeFieldBegin('section', TType.STRING, 37)
            oprot.writeString(self.section.encode('utf-8') if sys.version_info[0] == 2 else self.section)
            oprot.writeFieldEnd()
        if self.seriesName is not None:
            oprot.writeFieldBegin('seriesName', TType.STRING, 38)
            oprot.writeString(self.seriesName.encode('utf-8') if sys.version_info[0] == 2 else self.seriesName)
            oprot.writeFieldEnd()
        if self.slug is not None:
            oprot.writeFieldBegin('slug', TType.STRING, 39)
            oprot.writeString(self.slug.encode('utf-8') if sys.version_info[0] == 2 else self.slug)
            oprot.writeFieldEnd()
        if self.taxonomicClassifierList is not None:
            oprot.writeFieldBegin('taxonomicClassifierList', TType.LIST, 40)
            oprot.writeListBegin(TType.STRING, len(self.taxonomicClassifierList))
            for iter116 in self.taxonomicClassifierList:
                oprot.writeString(iter116.encode('utf-8') if sys.version_info[0] == 2 else iter116)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.titleList is not None:
            oprot.writeFieldBegin('titleList', TType.LIST, 41)
            oprot.writeListBegin(TType.STRING, len(self.titleList))
            for iter117 in self.titleList:
                oprot.writeString(iter117.encode('utf-8') if sys.version_info[0] == 2 else iter117)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.typesOfMaterialList is not None:
            oprot.writeFieldBegin('typesOfMaterialList', TType.LIST, 42)
            oprot.writeListBegin(TType.STRING, len(self.typesOfMaterialList))
            for iter118 in self.typesOfMaterialList:
                oprot.writeString(iter118.encode('utf-8') if sys.version_info[0] == 2 else iter118)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.url is not None:
            oprot.writeFieldBegin('url', TType.STRING, 43)
            oprot.writeString(self.url.encode('utf-8') if sys.version_info[0] == 2 else self.url)
            oprot.writeFieldEnd()
        if self.wordCount is not None:
            oprot.writeFieldBegin('wordCount', TType.I32, 44)
            oprot.writeI32(self.wordCount)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(NITFInfo)
NITFInfo.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'alternateURL', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'articleAbstract', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'authorBiography', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'banner', 'UTF8', None, ),  # 4
    (5, TType.LIST, 'biographicalCategoryList', (TType.STRING, 'UTF8', False), None, ),  # 5
    (6, TType.STRING, 'columnName', 'UTF8', None, ),  # 6
    (7, TType.I32, 'columnNumber', None, None, ),  # 7
    (8, TType.I64, 'correctionDate', None, None, ),  # 8
    (9, TType.STRING, 'correctionText', 'UTF8', None, ),  # 9
    (10, TType.STRING, 'credit', 'UTF8', None, ),  # 10
    (11, TType.STRING, 'dayOfWeek', 'UTF8', None, ),  # 11
    (12, TType.LIST, 'descriptorList', (TType.STRING, 'UTF8', False), None, ),  # 12
    (13, TType.STRING, 'featurePage', 'UTF8', None, ),  # 13
    (14, TType.LIST, 'generalOnlineDescriptorList', (TType.STRING, 'UTF8', False), None, ),  # 14
    (15, TType.I32, 'guid', None, None, ),  # 15
    (16, TType.STRING, 'kicker', 'UTF8', None, ),  # 16
    (17, TType.LIST, 'leadParagraphList', (TType.STRING, 'UTF8', False), None, ),  # 17
    (18, TType.LIST, 'locationList', (TType.STRING, 'UTF8', False), None, ),  # 18
    (19, TType.LIST, 'nameList', (TType.STRING, 'UTF8', False), None, ),  # 19
    (20, TType.STRING, 'newsDesk', 'UTF8', None, ),  # 20
    (21, TType.STRING, 'normalizedByline', 'UTF8', None, ),  # 21
    (22, TType.LIST, 'onlineDescriptorList', (TType.STRING, 'UTF8', False), None, ),  # 22
    (23, TType.STRING, 'onlineHeadline', 'UTF8', None, ),  # 23
    (24, TType.STRING, 'onlineLeadParagraph', 'UTF8', None, ),  # 24
    (25, TType.LIST, 'onlineLocationList', (TType.STRING, 'UTF8', False), None, ),  # 25
    (26, TType.LIST, 'onlineOrganizationList', (TType.STRING, 'UTF8', False), None, ),  # 26
    (27, TType.LIST, 'onlinePeople', (TType.STRING, 'UTF8', False), None, ),  # 27
    (28, TType.LIST, 'onlineSectionList', (TType.STRING, 'UTF8', False), None, ),  # 28
    (29, TType.LIST, 'onlineTitleList', (TType.STRING, 'UTF8', False), None, ),  # 29
    (30, TType.LIST, 'organizationList', (TType.STRING, 'UTF8', False), None, ),  # 30
    (31, TType.I32, 'page', None, None, ),  # 31
    (32, TType.LIST, 'peopleList', (TType.STRING, 'UTF8', False), None, ),  # 32
    (33, TType.I64, 'publicationDate', None, None, ),  # 33
    (34, TType.I32, 'publicationDayOfMonth', None, None, ),  # 34
    (35, TType.I32, 'publicationMonth', None, None, ),  # 35
    (36, TType.I32, 'publicationYear', None, None, ),  # 36
    (37, TType.STRING, 'section', 'UTF8', None, ),  # 37
    (38, TType.STRING, 'seriesName', 'UTF8', None, ),  # 38
    (39, TType.STRING, 'slug', 'UTF8', None, ),  # 39
    (40, TType.LIST, 'taxonomicClassifierList', (TType.STRING, 'UTF8', False), None, ),  # 40
    (41, TType.LIST, 'titleList', (TType.STRING, 'UTF8', False), None, ),  # 41
    (42, TType.LIST, 'typesOfMaterialList', (TType.STRING, 'UTF8', False), None, ),  # 42
    (43, TType.STRING, 'url', 'UTF8', None, ),  # 43
    (44, TType.I32, 'wordCount', None, None, ),  # 44
)
fix_spec(all_structs)
del all_structs



from ..annotate import AnnotateCommunicationService
from .service_wrapper import (
    ConcreteServiceClientWrapper,
    ConcreteServiceWrapper,
    HTTPConcreteServiceClientWrapper,
    SubprocessConcreteServiceWrapper)


AnnotateCommunicationClientWrapper = type(
    'AnnotateCommunicationClientWrapper',
    (ConcreteServiceClientWrapper,),
    {'concrete_service_class': AnnotateCommunicationService})


AnnotateCommunicationServiceWrapper = type(
    'AnnotateCommunicationServiceWrapper',
    (ConcreteServiceWrapper,),
    {'concrete_service_class': AnnotateCommunicationService})


HTTPAnnotateCommunicationClientWrapper = type(
    'HTTPAnnotateCommunicationClientWrapper',
    (HTTPConcreteServiceClientWrapper,),
    {'concrete_service_class': AnnotateCommunicationService})


SubprocessAnnotateCommunicationServiceWrapper = type(
    'SubprocessAnnotateCommunicationServiceWrapper',
    (SubprocessConcreteServiceWrapper,),
    {'concrete_service_wrapper_class': AnnotateCommunicationServiceWrapper})



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.services.Service
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(concrete.services.Service.Iface):
    """
    A service that exists so that clients can store Concrete data
    structures to implementing servers.

    Implement this if you are creating an analytic that wishes to
    store its results back to a server. That server may perform
    validation, write the new layers to a database, and so forth.
    """
    def store(self, communication):
        """
        Store a communication to a server implementing this method.

        The communication that is stored should contain the new
        analytic layers you wish to append. You may also wish to call
        methods that unset annotations you feel the receiver would not
        find useful in order to reduce network overhead.

        Parameters:
         - communication
        """
        pass


class Client(concrete.services.Service.Client, Iface):
    """
    A service that exists so that clients can store Concrete data
    structures to implementing servers.

    Implement this if you are creating an analytic that wishes to
    store its results back to a server. That server may perform
    validation, write the new layers to a database, and so forth.
    """
    def __init__(self, iprot, oprot=None):
        concrete.services.Service.Client.__init__(self, iprot, oprot)

    def store(self, communication):
        """
        Store a communication to a server implementing this method.

        The communication that is stored should contain the new
        analytic layers you wish to append. You may also wish to call
        methods that unset annotations you feel the receiver would not
        find useful in order to reduce network overhead.

        Parameters:
         - communication
        """
        self.send_store(communication)
        self.recv_store()

    def send_store(self, communication):
        self._oprot.writeMessageBegin('store', TMessageType.CALL, self._seqid)
        args = store_args()
        args.communication = communication
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_store(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = store_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.ex is not None:
            raise result.ex
        return


class Processor(concrete.services.Service.Processor, Iface, TProcessor):
    def __init__(self, handler):
        concrete.services.Service.Processor.__init__(self, handler)
        self._processMap["store"] = Processor.process_store

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_store(self, seqid, iprot, oprot):
        args = store_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = store_result()
        try:
            self._handler.store(args.communication)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except concrete.services.ttypes.ServicesException as ex:
            msg_type = TMessageType.REPLY
            result.ex = ex
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("store", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class store_args(object):
    """
    Attributes:
     - communication
    """


    def __init__(self, communication=None,):
        self.communication = communication

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.communication = concrete.communication.ttypes.Communication()
                    self.communication.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('store_args')
        if self.communication is not None:
            oprot.writeFieldBegin('communication', TType.STRUCT, 1)
            self.communication.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(store_args)
store_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'communication', [concrete.communication.ttypes.Communication, None], None, ),  # 1
)


class store_result(object):
    """
    Attributes:
     - ex
    """


    def __init__(self, ex=None,):
        self.ex = ex

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.ex = concrete.services.ttypes.ServicesException()
                    self.ex.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('store_result')
        if self.ex is not None:
            oprot.writeFieldBegin('ex', TType.STRUCT, 1)
            self.ex.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(store_result)
store_result.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'ex', [concrete.services.ttypes.ServicesException, None], None, ),  # 1
)
fix_spec(all_structs)
del all_structs




#!/usr/bin/env python

"""Print human-readable information about a Communication to stdout

concrete_inspect.py is a command-line script for printing out information
about a Concrete Communication.
"""
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import logging
import json
import sys

import concrete.version
import concrete.inspect
from concrete.util import (
    CommunicationReader, FileType, set_stdout_encoding,
    filter_annotations_json, lun
)


BASE_ANNOTATION_TYPES = (
    'id',
    'text',
    'sections',
    'entities',
    'mentions',
    'situations',
    'situation-mentions',
    'communication-taggings',
    'metadata',
    'treebank',
    'tokens',
    'dependency',
    'lemmas',
    'ner',
    'pos',
)


def print_header(header):
    print()
    print(header)
    print('-' * len(header))


def print_header_if(header, condition):
    if condition:
        print_header(header)


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description='''
Print information about a Concrete Communication to stdout.  If
communication_filename is specified, read communication from file;
otherwise, read from standard input.  One or more annotation flags must
be specified (calling this script with no arguments is an error).''',
        epilog='''
Filtering annotations:

    TYPE
        Should be one of the following:

            {}
            other-tag:TAG-TYPE

        (with TAG-TYPE replaced by a token tagging type given to the
        --other-tags flag).

    FILTER_ANNOTATIONS_JSON
        Can be for example

            {}

        to print annotations by the Goldenhorse tool or

            {}

        to print only the most recent annotation.

        Annotations of the specified type are filtered by parsing
        FILTER_ANNOTATIONS_JSON and passing the results as keyword
        arguments to concrete.util.metadata.filter_annotations, which
        has the following documentation at the time of writing (see
        the API documentation for updates):

        filter_fields (dict): dict of fields and their desired values
            by which to filter annotations (keep annotations whose
            field FIELD equals VALUE for all FIELD: VALUE entries).
            Default: keep all annotations.  Valid fields are:
            tool, timestamp, kBest.
        sort_field (str): field by which to re-order annotations.
            Default: do not re-order annotations.
        sort_reverse (bool): True to reverse order of annotations
            (after sorting, if any).
        action_if_multiple (str): action to take if, after filtering,
            there is more than one annotation left.  'pass' to
            return all filtered and re-ordered annotations, 'raise' to
            raise an exception of type MultipleAnnotationsError,
            'first' to return a list containing the first annotation
            after filtering and re-ordering, or 'last' to return a list
            containing the last annotation after filtering and
            re-ordering.
        action_if_zero (str): action to take if, after filtering, there
            are no annotations left.  'pass' to return an empty list,
            'raise' to raise an exception of type ZeroAnnotationsError.

        The default behavior is to do no filtering (or re-ordering).
        '''.format(
            '\n            '.join(BASE_ANNOTATION_TYPES),
            json.dumps(dict(filter_fields=dict(tool='Goldenhorse'))),
            json.dumps(dict(action_if_multiple='last', sort_field='timestamp')),
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--count', type=int,
                        help='Print at most this many communications.')
    parser.add_argument('--annotation-headers', action='store_true',
                        help='Print a header above each annotation '
                             '(useful if multiple annotation flags are '
                             'specified and it\'s not clear where the output '
                             'of one annotation ends and the output of the '
                             'next annotation begins)')
    parser.add_argument("--char-offsets",
                        help="Print token text extracted from character offset"
                             "s (not the text stored in the tokenization) in '"
                             "CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--starts",
                        help="Print starting indices (offsets) of tokens in "
                             "communication text, in 'CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--endings",
                        help="Print ending indices (offsets) of tokens in "
                             "communication text, in 'CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--dependency",
                        help="Print HEAD and DEPREL tags for first dependency "
                             "parse in 'CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--entities",
                        help="Print info about all Entities and their EntityMe"
                             "ntions",
                        action="store_true")
    parser.add_argument("--lemmas",
                        help="Print first set of lemma token tags in 'CoNLL-st"
                             "yle' format",
                        action="store_true")
    parser.add_argument("--metadata",
                        help="Print metadata for tools used to annotate Commun"
                             "ication",
                        action="store_true")
    parser.add_argument("--communication-taggings",
                        help="Print communication taggings",
                        action="store_true")
    parser.add_argument("--mentions",
                        help="Print whitespace-separated tokens, with entity m"
                             "entions wrapped using <ENTITY ID=x> tags, where "
                             "'x' is the (zero-indexed) entity number",
                        action="store_true")
    parser.add_argument("--ner",
                        help="Print first set of Named Entity Recognition toke"
                             "n tags in 'CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--pos",
                        help="Print first set of Part-Of-Speech token tags in "
                             "'CoNLL-style' format",
                        action="store_true")
    parser.add_argument("--sections",
                        action='store_true',
                        help="Print text according to Section offsets"
                             "(textSpan values). These textSpans are assumed "
                             "to be valid.")
    parser.add_argument("--situation-mentions",
                        help="Print info about all SituationMentions",
                        action="store_true")
    parser.add_argument("--situations",
                        help="Print info about all Situations and their Situat"
                             "ionMentions",
                        action="store_true")
    parser.add_argument("--other-tag",
                        metavar='TAG-TYPE',
                        help="Tagging type of other token tagging to display "
                             "(this flag can be specified multiple times to "
                             "display multiple other token taggings)",
                        action='append')
    parser.add_argument("--text",
                        help="Print .text field",
                        action="store_true")
    parser.add_argument("--tokens",
                        help="Print whitespace-seperated tokens for *all* Toke"
                             "nizations in a Communication.  Each sentence tok"
                             "enization is printed on a separate line, and "
                             "empty lines indicate a section break",
                        action="store_true")
    parser.add_argument("--treebank",
                        help="Print Penn-Treebank style parse trees for *all* "
                             "Constituent Parses in the Communication",
                        action="store_true")
    parser.add_argument("--id",
                        help='Print communication id',
                        action='store_true')
    parser.add_argument('--filter-annotations',
                        nargs=2,
                        metavar=('TYPE', 'FILTER_ANNOTATIONS_JSON'),
                        help='Filter (and/or re-order) annotations of '
                             'specified type TYPE according to '
                             'FILTER_ANNOTATIONS_JSON (a string containing '
                             'json-encoded arguments to the '
                             'filter_annotations function).  '
                             'See below for details.',
                        action='append')
    parser.add_argument("--no-references",
                        help="Don't add references to communication while "
                             "loading (may prevent some 'NoneType' errors but "
                             "is incompatible with some annotations)",
                        action="store_true")
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    parser.add_argument("--dependency-tool",
                        type=str,
                        help='Deprecated:  Filter --dependency output by tool.')
    parser.add_argument("--entities-tool",
                        type=str,
                        help='Deprecated:  Filter --entities output by tool.')
    parser.add_argument("--lemmas-tool",
                        type=str,
                        help='Deprecated:  Filter --lemmas output by tool.')
    parser.add_argument("--metadata-tool",
                        type=str,
                        help='Deprecated:  Filter --metadata output by tool.')
    parser.add_argument("--communication-taggings-tool",
                        type=str,
                        help='Deprecated:  Filter --communication-taggings output by tool.')
    parser.add_argument("--mentions-tool",
                        type=str,
                        help='Deprecated:  Filter --mentions output by tool.')
    parser.add_argument("--ner-tool",
                        type=str,
                        help='Deprecated:  Filter --ner output by tool.')
    parser.add_argument("--pos-tool",
                        type=str,
                        help='Deprecated:  Filter --pos output by tool.')
    parser.add_argument("--sections-tool",
                        type=str,
                        help='Deprecated:  Filter --sections output by tool.')
    parser.add_argument("--situation-mentions-tool",
                        type=str,
                        help='Deprecated:  Filter --situation-mentions output by tool.')
    parser.add_argument("--situations-tool",
                        type=str,
                        help='Deprecated:  Filter --situations output by tool.')
    parser.add_argument("--text-tool",
                        type=str,
                        help='Deprecated:  Filter --text output by tool.')
    parser.add_argument("--tokens-tool",
                        type=str,
                        help='Deprecated:  Filter --tokens output by tool.')
    parser.add_argument("--treebank-tool",
                        type=str,
                        help='Deprecated:  Filter --treebank output by tool.')
    parser.add_argument("--id-tool",
                        type=str,
                        help='Deprecated:  Filter --id output by tool.')
    parser.add_argument('communication_filename',
                        nargs='?',
                        type=str,
                        help='Path to a Concrete Communication from which '
                        'to display information. If not specified, read '
                        'from standard input')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    add_references = not args.no_references

    # if communication_filename is None, read from stdin
    if args.communication_filename is not None:
        comms = CommunicationReader(args.communication_filename,
                                    add_references=add_references)
    else:
        comms = CommunicationReader('/dev/fd/0', add_references=add_references,
                                    filetype=FileType.STREAM)

    # check that an annotation argument was specified
    if not (args.char_offsets or args.starts or args.endings or
            args.dependency or args.lemmas or args.ner or
            args.pos or args.entities or args.mentions or args.metadata or
            args.sections or args.situation_mentions or args.situations or
            args.text or args.tokens or args.treebank or args.id or
            args.communication_taggings or args.other_tag):
        parser.print_help()
        sys.exit(1)

    # show deprecation warnings for --x-tool
    if (args.dependency_tool or
            args.lemmas_tool or
            args.ner_tool or
            args.pos_tool or
            args.entities_tool or
            args.mentions_tool or
            args.metadata_tool or
            args.sections_tool or
            args.situation_mentions_tool or
            args.situations_tool or
            args.text_tool or
            args.tokens_tool or
            args.treebank_tool or
            args.id_tool or
            args.communication_taggings_tool):
        if args.filter_annotations:
            parser.print_help()
            sys.exit(1)
        else:
            logging.warning(
                '--*-tool is deprecated; please use --filter-annotations '
                'instead')

    # check that --x-tool was not specified without --x
    if ((args.dependency_tool and not args.dependency) or
            (args.lemmas_tool and not args.lemmas) or
            (args.ner_tool and not args.ner) or
            (args.pos_tool and not args.pos) or
            (args.entities_tool and not args.entities) or
            (args.mentions_tool and not args.mentions) or
            (args.metadata_tool and not args.metadata) or
            (args.sections_tool and not args.sections) or
            (args.situation_mentions_tool and not args.situation_mentions) or
            (args.situations_tool and not args.situations) or
            (args.text_tool and not args.text) or
            (args.tokens_tool and not args.tokens) or
            (args.treebank_tool and not args.treebank) or
            (args.id_tool and not args.id) or
            (args.communication_taggings_tool and
                not args.communication_taggings)):
        parser.print_help()
        sys.exit(1)

    # facilitate construction of filter functions from args
    filters_by_annotation_type = dict(lun(args.filter_annotations))

    def _get_annotation_filter(annotation_type):
        '''
        Return filter function for given annotation type, or None.
        '''
        if annotation_type in filters_by_annotation_type:
            return lambda annotations: filter_annotations_json(
                annotations, filters_by_annotation_type[annotation_type]
            )
        else:
            return None

    # loop over communications and print annotations
    comm_num = 0

    for (comm, _) in comms:
        if args.count is not None and comm_num == args.count:
            break

        if args.id:
            print_header_if('id', args.annotation_headers)
            concrete.inspect.print_id_for_communication(
                comm, tool=args.id_tool,
                communication_filter=_get_annotation_filter('id'))
        if args.text:
            print_header_if('text', args.annotation_headers)
            concrete.inspect.print_text_for_communication(
                comm, tool=args.text_tool,
                communication_filter=_get_annotation_filter('text'))
        if args.sections:
            print_header_if('sections', args.annotation_headers)
            concrete.inspect.print_sections(
                comm, tool=args.sections_tool,
                communication_filter=_get_annotation_filter('sections'))
        if args.tokens:
            print_header_if('tokens', args.annotation_headers)
            concrete.inspect.print_tokens_for_communication(
                comm, tool=args.tokens_tool,
                tokenization_filter=_get_annotation_filter('tokens'))
        if args.treebank:
            print_header_if('treebank', args.annotation_headers)
            concrete.inspect.print_penn_treebank_for_communication(
                comm, tool=args.treebank_tool,
                parse_filter=_get_annotation_filter('treebank'))
        if (args.char_offsets or args.starts or args.endings or
                args.dependency or args.lemmas or args.ner or
                args.pos or args.other_tag):
            print_header_if('conll', args.annotation_headers)
            concrete.inspect.print_conll_style_tags_for_communication(
                comm, char_offsets=args.char_offsets,
                starts=args.starts,
                endings=args.endings,
                dependency=args.dependency,
                lemmas=args.lemmas, ner=args.ner, pos=args.pos,
                other_tags=dict(
                    (t, _get_annotation_filter('other-tag:' + t))
                    for t in lun(args.other_tag)
                ),
                dependency_tool=args.dependency_tool,
                dependency_parse_filter=_get_annotation_filter('dependency'),
                lemmas_tool=args.lemmas_tool,
                lemmas_filter=_get_annotation_filter('lemmas'),
                pos_tool=args.pos_tool,
                pos_filter=_get_annotation_filter('pos'),
                ner_tool=args.ner_tool,
                ner_filter=_get_annotation_filter('ner'))
        if args.entities:
            print_header_if('entities', args.annotation_headers)
            concrete.inspect.print_entities(
                comm, tool=args.entities_tool,
                entity_set_filter=_get_annotation_filter('entities'))
        if args.mentions:
            print_header_if('mentions', args.annotation_headers)
            concrete.inspect.print_tokens_with_entityMentions(
                comm, tool=args.mentions_tool,
                entity_mention_set_filter=_get_annotation_filter('mentions'))
        if args.situations:
            print_header_if('situations', args.annotation_headers)
            concrete.inspect.print_situations(
                comm, tool=args.situations_tool,
                situation_set_filter=_get_annotation_filter('situations'))
        if args.situation_mentions:
            print_header_if('situation mentions', args.annotation_headers)
            concrete.inspect.print_situation_mentions(
                comm, tool=args.situation_mentions_tool,
                situation_mention_set_filter=_get_annotation_filter(
                    'situation-mentions'))
        if args.communication_taggings:
            print_header_if('communication taggings', args.annotation_headers)
            concrete.inspect.print_communication_taggings_for_communication(
                comm, tool=args.communication_taggings_tool,
                communication_tagging_filter=_get_annotation_filter(
                    'communication-taggings'))
        if args.metadata:
            print_header_if('metadata', args.annotation_headers)
            concrete.inspect.print_metadata(
                comm, tool=args.metadata_tool,
                annotation_filter=_get_annotation_filter('metadata'))

        comm_num += 1


if __name__ == "__main__":
    main()
